#!/usr/bin/env python
"""
    Automates the phasing process for a specified VCF file. The function
    allows users to select between the beagle (default) and shapeit 
    phasing algorithms.

    ##################
    Command-line Usage
    ##################
    The VCF file phaser may be called using the following command:

    .. code-block:: bash
        
        python vcf_phase.py

    *************
    Example usage
    *************
    Command-line to phase chromosome chr1 using Beagle:

    .. code-block:: bash
        
        python vcf_phase.py --vcf input.vcf --phase-algorithm beagle --phase-chr chr1

    Command-line to phase using SHAPEIT with 100 burn-in iterations:

    .. code-block:: bash
        
        python vcf_phase.py --vcf input.vcf.gz --phase-algorithm shapeit --shapeit-burn-iter 100

    ############
    Dependencies 
    ############
    * `SHAPEIT <https://mathgen.stats.ox.ac.uk/genetics_software/shapeit/shapeit.html>`_
    * `Beagle 5.0 <https://faculty.washington.edu/browning/beagle/beagle.html>`_
    * `BCFtools <https://samtools.github.io/bcftools/bcftools.html>`_ 
    * `plink 2.0 <https://www.cog-genomics.org/plink/2.0/>`_

    ############################
    Input Command-line Arguments
    ############################
    **--vcf** *<input_filename>*
        Argument used to define the filename of the VCF file to be phased.
    **--model-file** *<model_filename>*
        Argument used to define the model file. Please note that this argument cannot be 
        used with the **--pop-file** argument or individual-based filters.
    **--model** *<model_str>*
        Argument used to define the model (i.e. the individual(s) to include and/or the 
        populations for relevant statistics). May be used with any statistic. Please note 
        that this argument cannot be used with **--pop-file** argument or the 
        individual-based filters.

    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**.
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--out-format** *<vcf, vcf.gz, bcf>*
        Argument used to define the desired output format. Formats include: uncompressed 
        VCF (vcf); compressed VCF (vcf.gz) [default]; and BCF (bcf).
    **--overwrite**
        Argument used to define if previous output should be overwritten.
    
    ##############################
    Phasing Command-line Arguments
    ##############################
    **--phase-algorithm** *<beagle, shapeit>*
        Argument used to define the phasing algorithm. BEAGLE 5.0 (beagle) [default] and
        SHAPEIT (shapeit). Please note: Both algorithms possess algorithm-specific 
        arguments that may be found in their respective sections.
    **--Ne** *<Ne_int>*
        Argument used to define the effective population size.
    **--genetic-map** *<genetic_map_filename>*
        Argument used to define a genetic map file.
    **--phase-chr** *<chr>*
        Argument used to define a single chromosome to phase.
    **--phase-from-bp**
        Argument used to define the lower bound of positions to include. May only be 
        used with a single chromosome.
    **--phase-to-bp**
        Argument used to define the upper bound of positions to include. May only be 
        used with a single chromosome.
    **--random-seed** *<seed_int>*
        Argument used to define the seed value for the random number generator.

    **************************************
    SHAPEIT Phasing Command-line Arguments
    **************************************
    **--shapeit-burn-iter** *<iteration_int>*
        Argument used to define the number of burn-in iterations.
    **--shapeit-prune-iter** *<iteration_int>*
        Argument used to define the number of pruning iterations.
    **--shapeit-main-iter** *<iteration_int>*
        Argument used to define the number of main iterations.
    **--shapeit-states** *<state_int>*
        Argument used to define the number of conditioning states for haplotype 
        estimation.
    **--shapeit-window** *<Mb_float>*
        Argument used to define the model window size in Mb.

    *************************************
    BEAGLE Phasing Command-line Arguments
    *************************************
    **--beagle-burn-iter** *<iteration_int>*
        Argument used to define the number of burn-in iterations.
    **--beagle-iter** *<iteration_int>*
        Argument used to define the number of main iterations
    **--beagle-states** *<state_int>*
        Argument used to define the number of model states for genotype 
        estimation.
    **--beagle-error** *<probability>*
        Argument used to define the HMM allele mismatch probability.
    **--beagle-window** *<cM_float>*
        Argument used to define the sliding window size in cM.
    **--beagle-overlap** *<cM_float>*
        Argument used to define the overlap between neighboring windows in cM.
    **--beagle-step** *<cM_float>*
        Argument used to define the step length in cM used for identifying short 
        IBS segments.
    **--beagle-nsteps** *<windows_int>*
        Argument used to define the number of consecutive **--beagle-steps** used 
        for identifying long IBS segments.
    **--beagle-path** *<path>*
        Argument used to define the path to locate beagle.jar.    
"""

import os
import sys
import copy
import shutil
import argparse
import glob
import logging

# Import vcf format check function
from pgpipe.vcf_reader_func import checkFormat

# Import logging module
from pgpipe.logging_module import initLogger, logArgs

# Import model reader
from pgpipe.model import read_model_file

# Import beagle functions
from pgpipe.beagle import call_beagle, check_for_beagle_intermediate_files

# Import shapeit functions
from pgpipe.shapeit import call_shapeit, remove_shapeit_intermediate_files, check_for_shapeit_intermediate_files

# Import bcftools functions
from pgpipe.bcftools import get_unique_chrs, get_samples, chr_subset_file, concatenate, check_for_index, create_index

def phase_argument_parser(passed_arguments):
    '''
    VCF Phase Argument Parser

    Assign the parameters for VCF Phase using argparse.

    Parameters
    ----------
    passed_arguments : list, optional
        Parameters passed by another function. sys.argv is used if
        not given. 

    Raises
    ------
    IOError
        If the input, or other specified files do not exist
    '''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def parser_add_to_list ():
        '''Custom action to add items to a list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):

                # Clean up any commas
                value = [item.strip(',') for item in value]
                
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    phase_parser = argparse.ArgumentParser()

    # Input arguments
    phase_parser.add_argument('--vcf', help = "Defines the filename of the VCF", type = str, required = True, action = parser_confirm_file())

    # Output arguments
    phase_parser.add_argument('--out', help = 'Defines the complete output filename, overrides --out-prefix')
    phase_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', default = 'out')

    out_format_list = ['vcf', 'vcf.gz', 'bcf']
    out_format_default = 'vcf.gz'

    phase_parser.add_argument('--out-format', metavar = metavar_list(out_format_list), help = 'Defines the desired output format', type = str, choices = out_format_list, default = out_format_default)

    # Model file arguments
    phase_parser.add_argument('--model-file', help = 'Defines the model file', type = str, action = parser_confirm_file())
    phase_parser.add_argument('--model', help = 'Defines the model and the individual(s) to include', type = str)

    # Non-Model individual selection
    phase_parser.add_argument('--filter-include-indv', help = 'Defines the individual(s) to include. This argument may be used multiple times if desired', nargs = '+', type = str, action = parser_add_to_list())
    phase_parser.add_argument('--filter-exclude-indv', help = 'Defines the individual(s) to exclude. This argument may be used multiple times if desired', nargs = '+', type = str, action = parser_add_to_list())
    phase_parser.add_argument('--filter-include-indv-file', help = 'Defines a file of individuals to include', action = parser_confirm_file())
    phase_parser.add_argument('--filter-exclude-indv-file', help = 'Defines a file of individuals to exclude', action = parser_confirm_file())

    # General arguments
    phase_parser.add_argument('--overwrite', help = "Overwrite previous output files", action = 'store_true')
    phase_parser.add_argument('--beagle-path', help = "Defines the path to locate beagle.jar", type = str)

    # Galaxy Option to pipe log to stdout
    phase_parser.add_argument('--log-stdout', help = argparse.SUPPRESS, action = 'store_true')

    # Phase algorithm argument
    phasing_list = ['beagle', 'shapeit']
    phasing_default = 'beagle'
    phase_parser.add_argument('--phase-algorithm', metavar = metavar_list(phasing_list), help = 'Specifies the phase algorithm to be used', type = str, choices = phasing_list, default = phasing_default)

    # Common phasing arguments
    phase_parser.add_argument('--Ne', help = 'Defines the effective population size', type = int)
    phase_parser.add_argument('--random-seed', help="Defines the seed value for the random number generator", type = int)
    phase_parser.add_argument('--genetic-map', help = 'Genetic map filename', type = str, action = parser_confirm_file())
    phase_parser.add_argument('--phase-chr', help = 'Selects a single chromosome to phase', type = str)
    phase_parser.add_argument('--phase-from-bp', help = 'Lower bound of sites to include. May only be used with a single chromosome', type = int)
    phase_parser.add_argument('--phase-to-bp', help = 'Upper bound of sites to include. May only be used with a single chromosome', type = int)

    # Shapeit-specific options
    phase_parser.add_argument('--shapeit-burn-iter', help = 'Number of the burn-in iterations (shapeit)', type = int)
    phase_parser.add_argument('--shapeit-prune-iter', help = 'Number of pruning iterations (shapeit)', type = int)
    phase_parser.add_argument('--shapeit-main-iter', help = 'Number of main iterations (shapeit)', type = int)
    phase_parser.add_argument('--shapeit-states', help = 'Number of conditioning states for haplotype estimation (shapeit)', type = int)
    phase_parser.add_argument('--shapeit-window', help = 'Model window size in Mb (shapeit)', type = float)

    # Beagle-specific options
    phase_parser.add_argument('--beagle-burn-iter', help = 'Number of the burn-in iterations (beagle)', type = int)
    phase_parser.add_argument('--beagle-iter', help = 'Number of iterations after burn-in (beagle)', type = int)
    phase_parser.add_argument('--beagle-states', help = 'Number of model states for genotype estimation (beagle)', type = int)
    phase_parser.add_argument('--beagle-window', help = 'Sliding window size in cM (beagle)', type = float)
    phase_parser.add_argument('--beagle-overlap', help = 'Overlap between neighboring sliding windows in cM (beagle)', type = float)
    phase_parser.add_argument('--beagle-error', help = 'HMM allele mismatch probability (beagle)', type = float)
    phase_parser.add_argument('--beagle-step', help = 'Step length in cM used for identifying short IBS segments (beagle)', type = float)
    phase_parser.add_argument('--beagle-nsteps', help = 'Number of consecutive --beagle-steps used for identifying long IBS segments (beagle)', type = int)

    if passed_arguments:
        return phase_parser.parse_args(passed_arguments)
    else:
        return phase_parser.parse_args()

def log_to_stdout (log_filename):
    '''
        Write log file contents to stdout

        Used to allow galaxy users to see the contents of the log file. 
        Once completed, delete the log file.

        Raises
        ------
        IOError
            If a log file does not exist
    '''

    # Confirm the log file exists, raise an error if not
    if not os.path.isfile(log_filename):
        raise IOError('%s not found' % log_filename)

    # Open the log file
    log_file = open(log_filename)

    # Write the log file to stdout
    sys.stdout.write(log_file.readlines())

    # Close the log file
    log_file.close()

    # Delete the log file
    os.remove(log_filename)

def concatenate_logs (log_files, new_log_filename):
    '''
        Concatenates log files

        This function is used to concatenate log files. If a single chromosome
        is being analyzed, the function is used to concatenate the log files
        from multiple wrappers. If multiple chromosomes are being analyzed the
        function is used to concatenate the logs of each chromosome.

        Raises
        ------
        IOError
            If a log file does not exist
    '''

    # Open the new log file
    with open(new_log_filename, 'wb') as out_file:
        # Loop the logs to be combined
        for log_file in log_files:
            # Confirm the log file exists, raise an error if not
            if not os.path.isfile(log_file):
                raise IOError('%s not found' % log_file)
            # Open the log file
            with open(log_file, 'rb') as in_file:
                # Copy the contents of the file
                shutil.copyfileobj(in_file, out_file)

    # Remove old log files
    for log_file in log_files:
        os.remove(log_file)

def assign_vcf_extension (filename):
    # Checks if the file is unzipped, bgzipped, or gzipped
    vcfname_format = checkFormat(filename)

    if vcfname_format == 'vcf':
        return '.vcf'

    elif vcfname_format == 'gzip' or vcfname_format == 'bgzip':
        return '.vcf.gz'

    elif vcfname_format == 'bcf':
        raise Exception('BCF not supported in current version')

    else:
        raise Exception('Unknown file format')

def write_indv_file (filename, indv_list):

    # Open the individual file
    indv_file = open(filename, 'w')

    # Loop the list of individuals
    for indv in indv_list:

        # Write the individual to the file
        indv_file.write(indv + '\n')

    # Close the file
    indv_file.close()

def read_indv_file (filename):

    # List to hold individuals
    indv_list = []

    # Open the individual file
    with open(filename, 'r') as indv_file:

        # Read the individual lines
        for indv_line in indv_file:

            # Add the individual to the list
            indv_list.append(indv_line.strip())

    # Return the individuals
    return indv_list

def assign_filename_prefix (output_filename, output_format):

    '''
        Assigns a prefix using a filename

        Used to assign a unique prefix for jobs using the user specified
        filename. This is to avoid output files with the same prefix, 
        either from previous or ongoing jobs. This function is only used 
        if no prefix has been specified by the user.

        Parameters
        ----------
        output_filename : str
            Specifies the filename specified by the user
        output_format : str
            Specifies the file format suffix
        
        Returns
        -------
        unique_prefix: str
            Specifies the unqiue prefix

        Raises
        ------
        Exception
            If unable to assign a unique filename

    '''

    # Save the updated prefix
    updated_prefix = copy.deepcopy(output_filename)

    # Check if the file already exists 
    if os.path.isfile(updated_prefix + output_format):
        raise Exception('Unable to assign prefix output. %s already exists' % (updated_prefix + output_format))

    # Return the updated prefix
    return updated_prefix

def run (passed_arguments = []):
    '''
    Phaser for VCF files.

    This function uses the argparse-based function :py:func:`phase_argument_parser`
    to parse either sys.argv or passed_arguments to obtain the parameters below. 
    The parameters are then translated to their BEAGLE/SHAPEIT equivalent. Once all 
    the parameters are assigned, BEAGLE or SHAPEIT is called.

    Parameters
    ----------
    --vcf : str
       Filename of the VCF
    --out : str
        Complete output filename, overrides --out-prefix
    --out-prefix : str
        Output prefix
    --out-format : str
        Desired output format
    --model-file : str
        Model filename
    --model : str
        Model to use
    --phase-algorithm : str
        The phase algorithm to use
    --Ne : int
        Effective population size
    --random-seed : int
        Seed value for the random number generator
    --genetic-map : str
        Genetic map filename
    --phase-chr : str
        Chromosome to phase
    --phase-from-bp : int
        Lower bound of sites to include. Only usable with a single 
        chromosome
    --phase-to-bp : int
        Upper bound of sites to include. Only usable with a single 
        chromosome
    --shapeit-burn-iter : int
        Number of burn-in iterations
    --shapeit-prune-iter : int
        Number of pruning iterations
    --shapeit-main-iter : int
        Number of main iterations
    --shapeit-states : int
        Number of conditioning states for haplotype estimation
    --shapeit-window : float
        Model window size in Mb
    --beagle-burn-iter : int
        Number of the burn-in iterations
    --beagle-iter : int
        Number of main iterations
    --beagle-states : int
        Number of model states for genotype estimation
    --beagle-window : float
        Sliding window size in cM
    --beagle-overlap : float
        Overlap between neighboring sliding windows in cM
    --beagle-error : float
        HMM allele mismatch probability
    --beagle-step : float
        Step length in cM used for identifying short IBS segments
    --beagle-nsteps : int
        Number of consecutive --beagle-steps used for identifying 
        long IBS segments

    Raises
    ------
    IOError
        Output file already exists and --overwrite is not specified
    Exception
        Incompatible arguments
    '''

    # Grab VCF arguments from command line
    phase_args = phase_argument_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(phase_args, func_name = 'vcf_phase')

    # Assign file extension for VCF input file
    vcfname_ext = assign_vcf_extension(phase_args.vcf)

    logging.info('Input file assigned')

    # Used to confirm if the VCF input was renamed
    vcfname_renamed = False

    # Confirm input has correct file extension
    if vcfname_ext not in phase_args.vcf:

        # Confirm the file has been renamed
        vcfname_renamed = True

        # Rename the file
        os.rename(phase_args.vcf, phase_args.vcf + vcfname_ext)
        
        # Update the vcf argument with the updated name
        phase_args.vcf += vcfname_ext

    # List to hold beagle/shapeit arguments
    phase_call_args = []

    # bool to check if an eclude file was created
    exclude_file_created = False

    # bool to check if an include file was created
    include_file_created = False

    # String to hold filename of filter individuals, if created
    filter_indv_filename = ''

    # Check if the has specified the output filename, without a prefix
    if phase_args.out and '--out-prefix' not in passed_arguments and '--out-prefix' not in sys.argv:

        # Assign a prefix based on the output filename
        phase_args.out_prefix = assign_filename_prefix(phase_args.out, phase_args.out_format)

    # Assign expected phased output filename
    phased_output = '%s.%s' % (phase_args.out_prefix, phase_args.out_format)

    # Check if previous output should not be overwriten
    if not phase_args.overwrite:

        # Check if the user has specified an output filename
        if phase_args.out:

            # Check if previous output exists
            if os.path.exists(phase_args.out):
                raise IOError('Phased output already exists')

        else:

            # Check if previous output exists
            if os.path.exists(phased_output):
                raise IOError('Phased output already exists')

    # Performs actions related to --model-file and argument assignment
    if phase_args.model_file:

        # Check if a model has been specified
        if not phase_args.model:
            raise IOError('No selected model. Please use --model to select a model')

        # Read in the model file
        models_file = read_model_file(phase_args.model_file)

        # Check that the selected model was found in the file
        if phase_args.model not in models_file:
            raise IOError('Selected model "%s" not found in: %s' % (phase_args.model, phase_args.model_file))

        logging.info('Model assigned')

        if phase_args.phase_algorithm == 'beagle':

            # Get list of individuals to include
            inds_in_model = models_file[phase_args.model].inds

            # Create file with individuals to exclude (beagle has no include option)
            models_file.create_exclude_ind_file(inds_to_include = inds_in_model, overwrite = phase_args.overwrite)

            # Assign exclude file
            phase_call_args.append('excludesamples=' + models_file.exclude_file)

            # Confirm that the exclude file was created
            exclude_file_created = True

        elif phase_args.phase_algorithm == 'shapeit':

            # Assign the model
            selected_model = models_file[phase_args.model]

            # Create file with individuals to exclude (beagle has no include option)
            selected_model.create_ind_file(overwrite = phase_args.overwrite)

            # Assign exclude file
            phase_call_args.extend(['--include-ind', selected_model.ind_file])

            # Confirm that the include file was created
            include_file_created = True

    # Check if individuals to include or exclude were assigned
    elif phase_args.filter_include_indv or phase_args.filter_exclude_indv or phase_args.filter_include_indv_file or phase_args.filter_exclude_indv_file:

        # List of individuals to include
        include_indv_list = []

        # List of individuals to exclude
        exclude_indv_list = []

        # Check for individuals to include from the command-line
        if phase_args.filter_include_indv:

            # Add the individuals to the list
            include_indv_list.extend(phase_args.filter_include_indv)

        # Check for individuals to exclude from the command-line
        if phase_args.filter_exclude_indv:

            # Add the individuals to the list
            exclude_indv_list.extend(phase_args.filter_exclude_indv)

        # Check for individuals to include from a file
        if phase_args.filter_include_indv_file:

            # Read the individuals from the file
            include_file_indvs = read_indv_file(phase_args.filter_include_indv_file)

            # Add the individuals to the list
            include_indv_list.extend(include_file_indvs)

        # Check for individuals to exclude from a file
        if phase_args.filter_exclude_indv_file:

            # Read the individuals from the file
            exclude_file_indvs = read_indv_file(phase_args.filter_exclude_indv_file)

            # Add the individuals to the list
            exclude_indv_list.extend(exclude_file_indvs)

        # Assign the complete list of individuals within the file
        indvs_in_vcf = get_samples(phase_args.vcf)

        # Assign the individuals to include, checking both include/exclude lists
        indvs_to_include = [include_indv for include_indv in include_indv_list if include_indv not in exclude_indv_list]

        if phase_args.phase_algorithm == 'beagle':

            # Assign a filename for the indv file
            filter_indv_filename = 'filter_exclude_individuals'

            # Assign the individuals to exclude, using indvs_in_vcf and indvs_to_include 
            indvs_to_exclude = [indv_in_vcf for indv_in_vcf in indvs_in_vcf if indv_in_vcf not in indvs_to_include]

            # Create the indv file
            write_indv_file(filter_indv_filename, indvs_to_exclude)

            # Assign exclude file
            phase_call_args.append('excludesamples=' + filter_indv_filename)

        elif phase_args.phase_algorithm == 'shapeit':

            # Assign a filename for the indv file
            filter_indv_filename = 'filter_include_individuals'

            # Create the indv file
            write_indv_file(filter_indv_filename, indvs_to_include)

            # Assign exclude file
            phase_call_args.extend(['--include-ind', filter_indv_filename])

    # Get the list of chromosomes within the VCF
    chrs_in_vcf = get_unique_chrs(phase_args.vcf)

    # Check if the user specified chromosome is within the file
    if phase_args.phase_chr and (phase_args.phase_chr not in chrs_in_vcf):
        raise Exception('--phase-chr %s not found in %s' % (phase_args.phase_chr, phase_args.vcf))

    # Check if bp-based arguments are possible
    if (phase_args.phase_from_bp or phase_args.phase_to_bp) and (not phase_args.phase_chr and len(chrs_in_vcf) != 1):
        raise Exception('The --phase-from-bp and --phase-to-bp arguments require the --phase-chr '
                        'argument to function if multiple chrs are within %s' % phase_args.vcf)

    # Assign general arguments and call beagle
    if phase_args.phase_algorithm == 'beagle':

        if phase_args.beagle_path is not None:
            beagle_path = phase_args.beagle_path
        else:
            beagle_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),'bin')
        # Check for the presence of intermediate files from beagle
        check_for_beagle_intermediate_files(phase_args.out_prefix, phase_args.out_format, overwrite = phase_args.overwrite)

        # Assigns the input and output arguments
        phase_call_args.extend(['gt=' + phase_args.vcf,
                                'out=' + phase_args.out_prefix])

        # Assign the burn-in iter, if specified
        if phase_args.beagle_burn_iter:
            phase_call_args.append('burnin=' + str(phase_args.beagle_burn_iter))

        # Assign the iter, if specified
        if phase_args.beagle_iter:
            phase_call_args.append('iterations=' + str(phase_args.beagle_iter))

        # Assign the state count, if specified
        if phase_args.beagle_states:
            phase_call_args.append('phase-states=' + str(phase_args.beagle_states))

        # Assign the window length, if specified
        if phase_args.beagle_window:
            phase_call_args.append('window=' + str(phase_args.beagle_window))

        # Assign the window overlap, if specified
        if phase_args.beagle_overlap:
            phase_call_args.append('overlap=' + str(phase_args.beagle_overlap))

        # Assign the HMM error, if specified
        if phase_args.beagle_error:
            phase_call_args.append('err=' + str(phase_args.beagle_error))

        # Assign the step length, if specified
        if phase_args.beagle_step:
            phase_call_args.append('step=' + str(phase_args.beagle_step))

        # Assign the number of steps, if specified
        if phase_args.beagle_nsteps:
            phase_call_args.append('nsteps=' + str(phase_args.beagle_nsteps))

        # Assign the genetic map, if specified
        if phase_args.genetic_map:
            phase_call_args.append('map=' + phase_args.genetic_map)

        # Assign the effective pop size, if specified
        if phase_args.Ne:
            phase_call_args.append('ne=' + str(phase_args.Ne))

        # Assign the random seed, if specified
        if phase_args.random_seed:
            phase_call_args.append('seed=' + str(phase_args.random_seed))

        # Check if the file has a single chromosome
        if not phase_args.phase_chr and len(chrs_in_vcf) == 1:
                    
            # Assign the chromosome, as beagle requires it to be assigned
            phase_args.phase_chr = chrs_in_vcf[0]

            logging.info('Assigned chr %s for beagle' % chrs_in_vcf[0])

        # Assign the chromosome to phase, if specified
        if phase_args.phase_chr:

            # Store the chromosome argument
            chr_arg = 'chrom=%s' % phase_args.phase_chr

            # Check if either bp position arguments were specified
            if phase_args.phase_from_bp or phase_args.phase_to_bp:

                # List of the position arguments, in their required order
                position_args = [':', phase_args.phase_from_bp, '-', phase_args.phase_to_bp]

                # Filter the position arguments to remove empty values
                filttered_position_args = filter(None, position_args)

                # Map the arguments to str and add them to the chromosome argument
                chr_arg += ''.join(map(str, filttered_position_args))

            # Add the chromosome argument the argument list
            phase_call_args.append(chr_arg)

        # Call beagle wrapper
        call_beagle(beagle_path, list(map(str, phase_call_args)), phase_args.out_prefix, phase_args.out_format)

        # Rename log using phased_output
        shutil.move(phase_args.out_prefix + '.log', phased_output + '.log')

    # Assign general arguments and call shapeit
    elif phase_args.phase_algorithm == 'shapeit':

        # Assign the shapeit burn in iter, if specified
        if phase_args.shapeit_burn_iter:
            phase_call_args.extend(['--burn', str(phase_args.shapeit_burn_iter)])

        # Assign the shapeit prune iter, if specified
        if phase_args.shapeit_prune_iter:
            phase_call_args.extend(['--prune', str(phase_args.shapeit_prune_iter)])

        # Assign the shapeit main iter, if specified
        if phase_args.shapeit_main_iter:
            phase_call_args.extend(['--main', str(phase_args.shapeit_main_iter)])

        # Assign the number of shapeit states if specified
        if phase_args.shapeit_states:
            phase_call_args.extend(['--states', str(phase_args.shapeit_states)])

        # Assign the shapeit window size, if specified
        if phase_args.shapeit_window:
            phase_call_args.extend(['--window', str(phase_args.shapeit_window)])

        # Assign the genetic map, if specified
        if phase_args.genetic_map:
            phase_call_args.extend(['--input-map', phase_args.genetic_map])

        # Assign the effective pop size, if specified
        if phase_args.Ne:
            phase_call_args.extend(['--effective-size', str(phase_args.Ne)])

        # Assign the random seed, if specified
        if phase_args.random_seed:
            phase_call_args.extend(['--seed', str(phase_args.random_seed)])

        # Check if only a single shapeit run is required
        if phase_args.phase_chr or len(chrs_in_vcf) == 1:

            # Check for the presence of intermediate files from shapeit
            check_for_shapeit_intermediate_files(phase_args.out_prefix, overwrite = phase_args.overwrite)

            # Assign the default shapeit input filename
            shapeit_input_vcf = phase_args.vcf

            # Check if a single chromosome is selected
            if phase_args.phase_chr and len(chrs_in_vcf) > 1:

                # Assign the chromosome-specific input prefix
                chr_input_prefix = phase_args.vcf + '.' + phase_args.phase_chr

                # Assign the chr-specific file as the shapeit input filename
                shapeit_input_vcf = chr_input_prefix + '.vcf.gz'

                # Create the chromosome-specific input
                chr_subset_file(phase_args.vcf, phase_args.phase_chr, chr_input_prefix, 'vcf.gz', 
                                from_bp = phase_args.phase_from_bp, to_bp = phase_args.phase_to_bp, 
                                overwrite = phase_args.overwrite)

                logging.info('Chr %s subset created' % phase_args.phase_chr)

            # Assigns the input and output arguments for shapeit
            phase_call_args.extend(['--input-vcf', shapeit_input_vcf,
                                    '--output-max', phase_args.out_prefix,
                                    '--output-log', phase_args.out_prefix + '.phase.log'])

            # Assign the from bp position argument, if specified
            if phase_args.phase_from_bp:
                phase_call_args.extend(['--input-from', phase_args.phase_from_bp])

            # Assign the to bp position argument, if specified
            if phase_args.phase_to_bp:
                phase_call_args.extend(['--input-to', phase_args.phase_to_bp])

            # Call shapeit wrapper
            call_shapeit(list(map(str, phase_call_args)), phase_args.out_prefix, phase_args.out_format)

            # Combine the log files
            concatenate_logs([phase_args.out_prefix + '.phase.log', phase_args.out_prefix + '.log'],  phased_output + '.log')

            # Check if a chr subset file was created
            if phase_args.phase_chr and len(chrs_in_vcf) > 1:
                # Delete the chromosome-specific input
                os.remove(shapeit_input_vcf)

                logging.info('Chr %s subset deleted' % phase_args.phase_chr)

            # Remove intermediate files created by shapeit
            remove_shapeit_intermediate_files(phase_args.out_prefix)

        # Check if multiple shapeit runs are required
        else:

            # List to store the phased filenames
            phased_filename_list = []

            # List to store the phased logs
            phased_log_list = []

            logging.info('Multi-chr shapeit phasing assigned')

            for chr_in_vcf in chrs_in_vcf:

                logging.info('Chr %s assigned' % chr_in_vcf)

                # Copy the arguments for this run
                chr_call_args = copy.deepcopy(phase_call_args)

                # Assign the chromosome-specific output prefix
                chr_out_prefix = phase_args.out_prefix + '.' + chr_in_vcf

                # Check for the presence of intermediate files from shapeit
                check_for_shapeit_intermediate_files(chr_out_prefix)

                # Assign the expected chromosome-specific output filename
                chr_out_filename = '%s.%s' % (chr_out_prefix, phase_args.out_format)

                # Store the output filename
                phased_filename_list.append(chr_out_filename)

                # Assign the chromosome-specific input prefix
                chr_input_prefix = phase_args.vcf + '.' + chr_in_vcf

                # Assign the expected chromosome-specific input filename
                chr_input_filename = chr_input_prefix + '.vcf.gz'

                # Create the chromosome-specific input
                chr_subset_file(phase_args.vcf, chr_in_vcf, chr_input_prefix, 'vcf.gz', 
                                overwrite = phase_args.overwrite)

                # Assigns the input and output arguments for shapeit
                chr_call_args.extend(['--input-vcf', chr_input_filename,
                                      '--output-max', chr_out_prefix,
                                      '--output-log', chr_out_prefix + '.phase.log'])

                # Call shapeit wrapper
                call_shapeit(list(map(str, chr_call_args)), chr_out_prefix, phase_args.out_format)

                # Combine log files
                concatenate_logs([chr_out_prefix + '.phase.log', chr_out_prefix + '.log'],  chr_out_filename + '.log')

                # Store the filename of the combined logs
                phased_log_list.append(chr_out_filename + '.log')

                # Delete the chromosome-specific input
                os.remove(chr_input_filename)

                # Remove intermediate files created by shapeit
                remove_shapeit_intermediate_files(chr_out_prefix)

            # Concatenate the vcf files
            concatenate(phased_filename_list, phase_args.out_prefix, phase_args.out_format)

            logging.info('Concatenated chromosomes')

            # Combine the log files
            concatenate_logs(phased_log_list, phased_output + '.log')

    # Check if the log should be piped to the stdout
    if phase_args.log_stdout:

        # Write the log to stdout
        log_to_stdout(phased_output + '.log')

        # Rename output to phase_args.out, if specified
        if phase_args.out:
            shutil.move(phased_output, phase_args.out)

    # Check if log should be saved as a file
    else:

        # Rename output to phase_args.out, if specified
        if phase_args.out:
            shutil.move(phased_output, phase_args.out)
            shutil.move(phased_output + '.log', phase_args.out + '.log')

        logging.info('Log file created')

    # Reverts the VCF input file
    if vcfname_renamed:
        os.rename(phase_args.vcf, phase_args.vcf[:-len(vcfname_ext)])

    # Check if a filter indv file was created
    if filter_indv_filename:

        # Delete the file
        os.remove(filter_indv_filename)

if __name__ == "__main__":
    initLogger()
    run()



import os
import sys
import copy
import logging
import subprocess
import string
import random

#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

from pgpipe.vcf_reader_func import checkFormat
from pgpipe.bcftools import check_bcftools_for_errors

def log_vcftools_reference (out_filename, append_mode = False, ref_header = True):

    # Check if the file is to be written in append mode
    if append_mode:

        # Open the file
        log_file = open(out_filename + '.log', 'a')

    else:

        # Open the file
        log_file = open(out_filename + '.log', 'w')

    # Check if the ref header should be added
    if ref_header:
        log_file.write('Please Reference alongside the PPP:\n')

    log_file.write('Danecek, P. et al. The variant call format and VCFtools. '
                   'Bioinformatics 27, 2156-2158 (2011).')

    log_file.close()

    logging.info('Reference assigned')

def check_for_vcftools_output (vcftools_output):
    '''
        Checks for the previous vcftools output

        Confirms that neither a previous vcftools log or output file exists.

        Parameters
        ----------
        vcftools_output : str
            Specifies the output filename to be checked

        Raises
        ------
        IOError
            If the vcftools output file exists
        IOError
            If the vcftools log file exists

    '''
    # Check if output file already exists
    if os.path.isfile(vcftools_output):
        raise IOError('VCF output file already exists')

    logging.info('Output file assigned')

    # Check if log file already exists
    if os.path.isfile(vcftools_output + '.log'):
        raise IOError('Log file already exists')

    logging.info('Log file assigned')

def assign_vcftools_unique_prefix (output_prefix, output_format, random_seed = None, string_size = 6, string_limit = 10, assignment_limit = 100):

    '''
        Assigns a unique vcftools filename prefix

        Used to assign a unique filename prefix for vcftools jobs. If an
        output file with the same prefix and suffix, either from previous 
        or ongoing jobs, this function will generate a unique prefix. This 
        function is only used if no prefix has been specified by the user.

        Parameters
        ----------
        output_prefix : str
            Specifies the filename prefix
        output_format : str
            Specifies the file format suffix
        random_seed : int, str, optional
            Specifies the random seed
        string_size : int, optional
            Specifies the number of character in the random string
        string_limit  : int, optional
            Specifies the character limit of the random string
        assignment_limit : int, optional
            Specifies a limit the number of assignment attempts

        Returns
        -------
        updated_prefix: str
            Specifies the unqiue prefix


        Raises
        ------
        Exception
            If unable to assign a unique filename

    '''

    # Assign the random seed, with a string or int
    random.seed(random_seed)

    # Assign the assignment counter, used to avoid infinite loops
    assignment_counter = 0

    # Generate a random string
    random_string = ''.join(random.choice(string.ascii_uppercase + string.digits) for digit in range(string_size))

    # Holds the updated prefix that will be returned when unique
    updated_prefix = output_prefix + '.' + random_string

    # Loop until a unique filename is created
    while os.path.isfile(updated_prefix + output_format):

        # Add to the assignment counter
        assignment_counter += 1

        # Check if assignment counter has reached it's limit
        if assignment_counter > assignment_limit:

            # Set the assignment counter to zero
            assignment_counter = 0

            # Add to the string size
            string_size += 1

            # Check if string size has reached it's limit and report the error
            if string_size > string_limit:
                raise Exception('Unable to assign unique intermediate output')

        # Generate a random string
        random_string = ''.join(random.choice(string.ascii_uppercase + string.digits) for digit in range(string_size))

        # Holds the updated prefix that will be returned when unique
        updated_prefix = output_prefix + '.' + random_string

    # Return the intermediate output filename
    return updated_prefix

def assign_vcftools_filename_prefix (output_prefix, output_format, output_filename):

    '''
        Assigns a vcftools prefix using a filename

        Used to assign a unique prefix for vcftools jobs using the user
        specified filename. This is to avoid output file with the same 
        prefix, either from previous or ongoing jobs. This function is 
        only used if no prefix has been specified by the user.

        Parameters
        ----------
        output_prefix : str
            Specifies the filename prefix
        output_format : str
            Specifies the file format suffix
        output_filename : str
            Specifies the filename specified by the user

        Returns
        -------
        unique_prefix: str
            Specifies the unqiue prefix

        Raises
        ------
        Exception
            If unable to assign a unique filename

    '''

    # List to hold the complete file path
    file_path_list = []

    # Split the filename
    split_file_path = os.path.split(output_filename)

    while split_file_path[1]:

        # Add the split path section to the file path list
        file_path_list = [split_file_path[1]] + file_path_list

        # Split the filename
        split_file_path = os.path.split(split_file_path[0])

    # Save the updated prefix
    updated_prefix = ''.join(file_path_list).replace('.','')

    # Check if the file already exists 
    if os.path.isfile(updated_prefix + output_format):
        raise Exception('Unable to assign prefix output. %s already exists' % (updated_prefix + output_format))

    return updated_prefix

def check_vcftools_for_errors (vcftools_stderr):
    '''
        Checks the vcftools stderr for errors

        Parameters
        ----------
        vcftools_stderr : str
            vcftools stderr

        Raises
        ------
        IOError
            If vcftools stderr returns an error
    '''

    # Returns True if the job completed without error
    if 'Run Time' in str(vcftools_stderr):
        pass

    # Print output for vcftools if error is detected
    elif 'Error' in str(vcftools_stderr):
        # Splits log into list of lines
        vcftools_stderr_lines = vcftools_stderr.splitlines()
        # Prints the error(s)
        raise Exception('\n'.join((output_line for output_line in vcftools_stderr_lines if output_line.startswith('Error'))))

    # Print output if not completed and no error found. Unlikely to be used, but included.
    else:
        raise Exception(vcftools_stderr)

def produce_vcftools_output (output, filename, append_mode = False, strip_header = False):
    '''
        Creates the vcftools output file

        This function will create an output file from the vcftools stdout.
        Please run `check_vcftools_for_errors` prior to check that vcftools
        finished without error.

        Parameters
        ----------
        output : str
            vcftools stdout
        filename : str
            Specifies the filename for the output file
        append_mode : bool
            Used to create a single output file from multiple calls
        strip_header : bool
            Used to remove the header if not needed

        Returns
        -------
        output : file
            vcftools output file

    '''

    # Check if the header should be stripped
    if strip_header:
        output = ''.join(output.splitlines(True)[1:])

    # Check if single log file is required from multiple calls
    if append_mode:
        vcftools_log_file = open(filename,'a')
    else:
        vcftools_log_file = open(filename,'w')

    vcftools_log_file.write(str(output))
    vcftools_log_file.close()

def produce_vcftools_log (output, filename, append_mode = False):
    '''
        Creates the vcftools log file

        This function will create a log file from the vcftools stderr. Please
        run `check_vcftools_for_errors` prior to check that vcftools finished
        without error.

        Parameters
        ----------
        output : str
            vcftools stderr
        filename : str
            Specifies the filename for the log file
        append_mode : bool
            Used to create a single log file from multiple calls

        Returns
        -------
        output : file
            vcftools log file

    '''
    # Check if single log file is required from multiple calls
    if append_mode:
        vcftools_log_file = open(filename + '.log','a')
    else:
        vcftools_log_file = open(filename + '.log','w')

    vcftools_log_file.write(str(output))
    vcftools_log_file.close()

def assign_vcftools_input_arg (filename):
    '''
        Confirms file format for vcftools

        Parameters
        ----------
        filename : str
            Specifies the input filename of unknown format

        Returns
        -------
        list
            Returns vcftools input command for `filename`

        Raises
        ------
        IOError
            If filename is an unknown file format
    '''

    # True if file extensions is recognized by vcftools
    if filename.endswith('.vcf') or filename.endswith('.vcf.gz') or filename.endswith('.bcf'):
        # Assign the associated input command
        if filename.endswith('.vcf'):
            return ['--vcf', filename]
        elif filename.endswith('.vcf.gz'):
            return ['--gzvcf', filename]
        elif filename.endswith('.bcf'):
            return ['--bcf', filename]

    # True if file extension is unknown or not recognized
    else:

        # Checks if the file is unzipped, bgzipped, or gzipped
        vcfname_format = checkFormat(filename)

        # Assign the associated input command, or return an error.
        if vcfname_format == 'vcf':
            return ['--vcf', filename]
        elif vcfname_format == 'bgzip':
            return ['--gzvcf', filename]
        elif vcfname_format == 'bcf':
            return ['--bcf', filename]
        else:
            raise Exception('Unknown VCF file format')

def check_bgzip_for_errors (bgzip_stderr):
    '''
        Checks the bgzip stderr for errors

        Parameters
        ----------
        bgzip_stderr : str
            bgzip stderr

        Raises
        ------
        IOError
            If bgzip stderr returns an error
    '''

    if bgzip_stderr:
        raise IOError('Error occured while compressing the vcf file')

def bgzip_decompress_vcfgz (vcfgz_filename, out_prefix = None, keep_original = False):
        '''
            Converts a vcf.gz to vcf

            The function automates bgzip to decompress a vcf.gz file into a vcf

            Parameters
            ----------
            vcfgz_filename : str
                The file name of the vcf.gz file to be decompressed
            out_prefix : str
                Output file prefix (i.e. filename without extension)
            keep_original : bool
                Specifies if the original file should be kept

            Raises
            ------
            IOError
                Error in creating the compressed file
        '''

        # Run bgzip with stdout piped to file
        if keep_original or out_prefix:

            if out_prefix:

                # Assign the bgzip filename
                vcf_filename = out_prefix + '.vcf'

            else:

                # Seperate into path and filename
                split_path, split_filename = os.path.split(vcfgz_filename)

                # Remove any file extensions
                vcf_basename = split_filename.split(os.extsep)[0] + '.vcf'

                # Join path and filename
                vcf_filename = os.path.join(split_path, vcf_basename)

            # Create the output file
            vcf_file = open(vcf_filename, 'w')

            # bgzip subprocess call
            bgzip_call = subprocess.Popen(['bgzip', '-dc', vcfgz_filename], stdout = vcf_file, stderr = subprocess.PIPE)

        # Run bgzip normally
        else:

            # bgzip subprocess call
            bgzip_call = subprocess.Popen(['bgzip', '-d', vcfgz_filename], stdout = subprocess.PIPE, stderr = subprocess.PIPE)

        # Save the stdout and stderr from bgzip
        bgzip_out, bgzip_err = bgzip_call.communicate()

        # Check that output file was compressed correctly
        check_bgzip_for_errors(bgzip_err)

        # Delete input when also using an output prefix
        if out_prefix and not keep_original:
            os.remove(vcfgz_filename)

def bgzip_compress_vcf (vcf_filename, out_prefix = '', keep_original = False):
        '''
            Converts a vcf to vcf.gz

            The function automates bgzip to compress a vcf file into a vcf.gz

            Parameters
            ----------
            vcf_filename : str
                The file name of the vcf file to be compressed
            keep_original : bool
                Specifies if the original file should be kept

            Raises
            ------
            IOError
                Error in creating the compressed file
        '''

        # Compress and keep the original file
        if keep_original or out_prefix:

            if out_prefix:

                # Assign the filename
                vcfgz_filename = out_prefix + '.vcf.gz'

            else:

                # Seperate into path and filename
                split_path, split_filename = os.path.split(vcfgz_filename)

                # Remove any file extensions
                vcfgz_basename = split_filename.split(os.extsep)[0] + '.vcf.gz'

                # Join path and filename
                vcfgz_filename = os.path.join(split_path, vcfgz_basename)


            # Create the output file
            vcfgz_file = open(vcfgz_filename, 'w')

            # bgzip subprocess call
            bgzip_call = subprocess.Popen(['bgzip', '-c', vcf_filename], stdout = vcfgz_file, stderr = subprocess.PIPE)

        # Compress and do not keep the original file
        else:

            # bgzip subprocess call
            bgzip_call = subprocess.Popen(['bgzip', vcf_filename], stdout = subprocess.PIPE, stderr = subprocess.PIPE)

        # Save the stdout and stderr from bgzip
        bgzip_out, bgzip_err = bgzip_call.communicate()

        # Check that output file was compressed correctly
        check_bgzip_for_errors(bgzip_err)

def cvt_vcftools_site_to_bed (vcftools_out_str):
    # Check if str in the header
    if 'CHROM' not in vcftools_out_str or 'POS' not in vcftools_out_str:
        # Split the line into a list
        vcftools_out_data = vcftools_out_str.strip().split('\t')
        # Calc chromEnd
        chrom_end = copy.deepcopy(vcftools_out_data[1])
        # Convert the chromStart to chromStart0
        vcftools_out_data[1] = int(vcftools_out_data[1]) - 1
        # Add chrom_end to the list
        vcftools_out_data = vcftools_out_data + [chrom_end]
        # Return the list as a string (with newline element)
        return '\t'.join(map(str, vcftools_out_data)) + '\n'
    else:
        # Remove the header
        return ''

def pipe_vcftools_to_bed_file (vcftools_call_args, output_filename):

    '''
        Pipes site-file output of vcftools to a bed formmated file

        The purpose of this function is to avoid creating large uncompressed
        vcf files by directly piping the output of vcftools to bgzip. This
        results in creating a vcf.gz file without any intermediates.

        Parameters
        ----------
        vcftools_call_args : list
            vcftools arguments
        output_filename : str
            Filename of the bed file

    '''

    # Open vcftools pipe
    vcftools_call = subprocess.Popen(['vcftools', '--stdout'] + list(map(str, vcftools_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Create the bed file
    bed_output = open(output_filename, 'w')

    try:
        # Iterate the vcftools stdout unless error occurs
        for vcftools_stdout_line in iter(vcftools_call.stdout.readline, b''):
            bed_output.write(cvt_vcftools_site_to_bed(vcftools_stdout_line))
        # Close the bed file
        bed_output.close()
    except:
        # Close the bed file
        bed_output.close()
        # Delete the file
        os.remove(output_filename)

    # Wait for vctools to finish
    vcftools_call.wait()

    # Close the vcftools stdout
    vcftools_call.stdout.close()

    # Read the vcftools stderr
    vcftools_stderr = vcftools_call.stderr.read()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        vcftools_stderr = vcftools_stderr.decode()

    # Check that the log file was created correctly
    check_vcftools_for_errors(vcftools_stderr)

    logging.info('vcftools call complete')

    return vcftools_stderr

def pipe_vcftools_bgzip (vcftools_call_args, output_filename):
    '''
        Pipes the output of vcftools to bgzip

        The purpose of this function is to avoid creating large uncompressed
        vcf files by directly piping the output of vcftools to bgzip. This
        results in creating a vcf.gz file without any intermediates.

        Parameters
        ----------
        vcftools_call_args : list
            vcftools arguments
        output_filename : str
            Filename of the compressed vcf file

    '''

    vcftools_call = subprocess.Popen(['vcftools', '--stdout'] + list(map(str, vcftools_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Create bgzip output file
    bgzip_output = open(output_filename, 'wb')

    # bgzip subprocess call
    bgzip_call = subprocess.Popen(['bgzip'], stdin = vcftools_call.stdout, stdout = bgzip_output, stderr = subprocess.PIPE)

    # Wait for vctools to finish
    vcftools_call.wait()

    # Close the vcftools stdout
    vcftools_call.stdout.close()

    # Read the vcftools stderr
    vcftools_stderr = vcftools_call.stderr.read()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        vcftools_stderr = vcftools_stderr.decode()

    # Check that the log file was created correctly
    check_vcftools_for_errors(vcftools_stderr)

    # Wait for bgzip to finish
    bgzip_call.wait()

    # Close the compressed vcf file
    bgzip_output.close()

    # Save the stderr from bgzip, stdout = None
    bgzip_stdout, bgzip_stderr = bgzip_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        bgzip_stderr = bgzip_stderr.decode()

    # Check that output file was compressed correctly
    check_bgzip_for_errors(bgzip_stderr)

    logging.info('vcftools and bgzip calls complete')

    return vcftools_stderr

def pipe_vcftools_bcftools (vcftools_call_args, output_filename):
    '''
        Pipes the output of vcftools to bcftools

        The purpose of this function is to avoid the vcftools command
        --recode-bcf that may result in malformed BCF files. To avoid large
        uncompressed intermediates, this function pipes the stdout of vcftools
        to bcftools.

        Parameters
        ----------
        vcftools_call_args : list
            vcftools arguments
        output_filename : str
            Filename of the BCF file

    '''

    vcftools_call = subprocess.Popen(['vcftools', '--stdout'] + list(map(str, vcftools_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Holds the arguments to convert to BCF format
    convert_args = ['view', '-O', 'b']

    # Create bgzip output file
    bcftools_output = open(output_filename, 'wb')

    # bcftools subprocess call
    bcftools_call = subprocess.Popen(['bcftools'] + convert_args, stdin = vcftools_call.stdout, stdout = bcftools_output, stderr = subprocess.PIPE)

    # Wait for vctools to finish
    vcftools_call.wait()

    # Close the vcftools stdout
    vcftools_call.stdout.close()

    # Read the vcftools stderr
    vcftools_stderr = vcftools_call.stderr.read()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        vcftools_stderr = vcftools_stderr.decode()

    # Check that the log file was created correctly
    check_vcftools_for_errors(vcftools_stderr)

    # Wait for bgzip to finish
    bcftools_call.wait()

    # Save the stderr from bgzip, stdout = None
    bcftools_stdout, bcftools_stderr = bcftools_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        bcftools_stderr = bcftools_stderr.decode()

    # Check that output file was compressed correctly
    check_bcftools_for_errors(bcftools_stderr)

    logging.info('vcftools and bcftools calls complete')

    return vcftools_stderr

def pipe_vcftools_to_file (vcftools_call_args, output_filename, append_output = False):
    '''
        Pipes file output of vcftools to a standard file

        The function calls vcftools. Returns the stderr of vcftools to
        create log file of the call. The function may be used to append multiple
        calls to vcftools to a single file

        Parameters
        ----------
        vcftools_call_args : list
            vcftools arguments
        append_output : bool
            The output file should be written in append mode

        Returns
        -------
        vcftools_err : str
            vcftools log output

        Raises
        ------
        Exception
            If vcftools stderr returns an error
    '''

    # Open vcftools pipe
    vcftools_call = subprocess.Popen(['vcftools', '--stdout'] + list(map(str, vcftools_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Check if the output should be opened in append mode
    if append_output:
        # Create the output file (in append mode)
        output_file = open(output_filename, 'a')
    else:
        # Create the output file (in write mode)
        output_file = open(output_filename, 'w')

    try:
        # Create iterator of the vcftools stdout
        stdout_iter = iter(vcftools_call.stdout.readline, b'')

        # Check if the output is being appended and the file is empty
        if append_output and os.stat(output_filename).st_size != 0:
            # Skip the header if the file isn't empty and appending
            next(stdout_iter)

        # Iterate the vcftools stdout
        for vcftools_stdout_line in stdout_iter:

            # Check if code is running in python 3
            if sys.version_info[0] == 3:
                # Convert bytes to string
                vcftools_stdout_line = vcftools_stdout_line.decode()

            output_file.write(vcftools_stdout_line)

        # Close the bed file
        output_file.close()

    except:
        # Close the bed file
        output_file.close()
        # Delete the file
        os.remove(output_filename)

        raise Exception('vcftools to python pipe error')

    # Wait for vctools to finish
    vcftools_call.wait()

    # Close the vcftools stdout
    vcftools_call.stdout.close()

    # Read the vcftools stderr
    vcftools_stderr = vcftools_call.stderr.read()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        vcftools_stderr = vcftools_stderr.decode()

    # Check that the log file was created correctly
    check_vcftools_for_errors(vcftools_stderr)

    logging.info('vcftools call complete')

    return vcftools_stderr

def standard_vcftools_call (vcftools_call_args):
    '''
        Calls vcftools

        The function calls vcftools. Returns the stderr of vcftools to
        create log file of the call.

        Parameters
        ----------
        vcftools_call_args : list
            vcftools arguments

        Returns
        -------
        vcftools_out : str
            vcftools call output
        vcftools_err : str
            vcftools log output

        Raises
        ------
        Exception
            If vcftools stderr returns an error
    '''

    # vcftools subprocess call without stdout
    vcftools_call = subprocess.Popen(['vcftools'] + list(map(str, vcftools_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Wait for vcftools to finish
    vcftools_stdout, vcftools_stderr = vcftools_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        vcftools_stderr = vcftools_stderr.decode()

    logging.info('vcftools call complete')

    # Check that the log file was created correctly
    check_vcftools_for_errors(vcftools_stderr)

    return vcftools_stderr

def call_vcftools (vcftools_call_args, output_format = None, output_filename = None, append_mode = False):
    '''
        Calls vcftools

        The function calls vcftools. Returns the stderr of vcftools to
        create log file of the call.

        Parameters
        ----------
        vcftools_call_args : list
            vcftools arguments
        output_format : str, optional
            Allows for specific formats to be given which may result in
            the stdout of vcftools being piped to another script or 
            program
        output_filename : str, optional
            The output filename to be assigned if the stdout of vcftools 
            is piped
        append_mode : bool, optional
            Allows for vcftools stdout to append a file. Currently 
            incompatible with output_format

        Returns
        -------
        vcftools_err : str
            vcftools log output

        Raises
        ------
        Exception
            If vcftools stderr returns an error
    ''' 

    # Check if an output format was specified
    if output_format and output_filename: 

        # Check if the output is a bgzipped vcf
        if output_format == 'vcf.gz':

            # Pipe vcftools stdout to bgzip to create a bgzipped vcf
            vcftools_err = pipe_vcftools_bgzip(vcftools_call_args, output_filename)

        # Check if the output is a bcf
        elif output_format == 'bcf':

            # Pipe vcftools stdout to bgzip to create a bgzipped vcf
            vcftools_err = pipe_vcftools_bcftools(vcftools_call_args, output_filename)

        # Check if the output is a bed-based file
        elif output_format in ['removed_bed','kept_bed']:

            # Pipe vcftools stdout to bed file
            vcftools_err = pipe_vcftools_to_bed_file(vcftools_call_args, output_filename)

        # Check if the output is another format, that does not require a pipe
        else:
            
            # Pipe the vcftools stdout to a standard file
            vcftools_err = pipe_vcftools_to_file(vcftools_call_args, output_filename)

    # Check if a filename but no format was specified 
    elif output_filename:

        # Check if the output should be written in append mode
        if append_mode:

            # Pipe the vcftools stdout to a standard file and allow appending
            vcftools_err = pipe_vcftools_to_file(vcftools_call_args, output_filename, append_output = True)

        else:

            # Pipe the vcftools stdout to a standard file
            vcftools_err = pipe_vcftools_to_file(vcftools_call_args, output_filename)

    # If no format or filename is given, run vcftools with just the passed arguments
    else:

        # Call vcftools under standard conditions, if not 
        vcftools_err = standard_vcftools_call(vcftools_call_args)

    # Return the log
    return vcftools_err




import sys
import argparse

from pgpipe.vcf_reader_func import getAlleleStats, VcfReader, getAlleleCountDict, getAlleleStatsTwo
from pgpipe.genome_region import RegionList, Region

class recordCache():
    def __init__(self):
        self.chrom = ''
        self.cache = {}

    def getSiteStat(self,record):
        return piSite(record)

    def getWindowStat(self,recgen,region):
        return piWindow(recgen,region)

def createParser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--vcf",dest="vcf",type=str)
    parser.add_argument("--window-stat",dest="window",action="store_true")
    parser.add_argument("--window-size",dest="window_size",type=int)
    parser.add_argument("--step-size",dest="step_size",type=int)
    parser.add_argument("--bed",dest="bedfile",type=str)
    parser.add_argument("--subsamp",dest="subsamp",type=str)
    return parser

def piSite(record):
    #alleles,total_sites,missing_inds = getAlleleStats(record)
    alleles,total_sites,missing_inds = getAlleleStatsTwo(record)
    #return 2
    site_sum = 0
    for a in record.alleles:
        #print (a)
        a_count = alleles[a]
        site_sum += (a_count)*(total_sites-a_count)
    return float(site_sum)/(float(total_sites)*float(total_sites-1))

def getCacheVals(cache,region):
    res_list = []
    for pos in list(cache.keys()):
        if int(pos) < region.start:
            del cache[pos]
        elif int(pos) < region.end:
            res_list.append(cache[pos])
    return res_list


def piWindow(recgen,region,cache=None):
    pi_sum = 0
    if cache is not None:
        cache_list = getCacheVals(cache,region)
        for cv in cache_list:
            pi_sum += cv
    for rec in recgen:
        ps = piSite(rec)
        pi_sum += ps
        if cache is not None:
            cache[str(rec.pos)] = ps
    if pi_sum == 0:
        return -1
    return pi_sum/float(len(region))



parser = createParser()
args = parser.parse_args()
if args.step_size is not None and args.window_size is None:
    raise Exception("--step-size requires --window-size")
if args.step_size is None and args.window_size is not None:
    args.step_size = args.window_size

idx_list = None
if args.subsamp is not None:
    idx_list = [int(i.strip()) for i in open(args.subsamp).readlines()]
cache = None
if args.window_size > args.step_size:
    cache = {}
#vcfname = str(sys.argv[1])
reglist = None
#if len(sys.argv) == 3:
if args.bedfile is not None:
    regname = str(sys.argv[2])
    reglist = RegionList(filename=args.bedfile,zeroho=True)

vcf_reader = VcfReader(args.vcf)

if reglist is None:
    #region = None
    regcount = 1
else:
    regcount = len(reglist.regions)

for i in range(regcount):
    if reglist is None:
        region = None
    else:
        region = reglist.regions[i]
    if args.window:
        if args.window_size is not None:
            eof = False
            if region is None:
                current_chrom = vcf_reader.info_rec.chrom
                subregion = Region(0,args.window_size,current_chrom)
            else:
                region_size = len(region)
                subregion = Region(region.start,min(region.start+args.window_size,region.end),region.chrom)
            while True:
                pw = piWindow(vcf_reader.getRegionIter(region=subregion),subregion,cache)
                if pw > -.5:
                    sys.stdout.write(subregion.toStr(zeroho=True)+'\t'+str(pw)+'\n')
                if region is None:
                    if vcf_reader.prev_last_rec is None:
                        eof = True
                        break
                    if vcf_reader.prev_last_rec.chrom != current_chrom:
                        current_chrom = vcf_reader.prev_last_rec.chrom
                        subregion = Region(0,args.window_size,current_chrom)
                    else:
                        subregion.start += args.step_size
                        subregion.end += args.step_size
                else:
                    if vcf_reader.prev_last_rec is None:
                        eof = True
                        break
                    if vcf_reader.prev_last_rec.chrom != region.chrom:
                        break
                    else:
                        subregion.start += args.step_size
                        subregion.end = min(subregion.end+args.step_size,region.end)
                        if subregion.start > region.end:
                            break
            

        else:
            pw = piWindow(vcf_reader.getRegionIter(region=region),region,cache)
            if pw > -.5:
                sys.stdout.write(region.toStr(zeroho=True)+'\t'+str(pw)+'\n')
    else:
        for rec in vcf_reader.getRegionIter(region=region):
            sys.stdout.write(str(rec.pos)+'\t'+str(piSite(rec))+'\n')




'''
    Given a BED file, return a BED file that has regions outside of the 
    regions in the input.

    This function operates similarly to 'bedtools complement', however it 
    does not require a file that contains chromosome lengths, as it will 
    automatically discard any bases after the last region on a chromosome
    in the input file. Options are included to extend the regions in the
    input (hence shrinking regions in the output), formatting for interval
    format, indicating what columns represent start/end/chromosome data,
    and adding/removing the 'chr' prefix from chromosome names.

    ###############
    Input Arguments
    ###############
    **--bed** *<input_bed>*
        Input BED filename
    
    ################
    Output Arguments
    ################
    **--out** *<output_filename>*
        Name of output file

    ################
    Region Arguments
    ################
    **--bed-column-index** *<start_idx>*,*<end_idx>*,*<chrom_idx>*
        Comma-separated string of the zero-based indices of the start, end,
        and chromosome columns in the input file, so the file doesn't need to
        be reformatted. Default for a regular BED file is 1,2,0.
    **--zero-ho**
        If set, indicates input BED regions are on zero-based, half-open
        coordinate system, as opposed to one-based, closed. For example, the
        first million bases on a chromosome would be:
            Zero-based, half-open: 0,1000000
            One-based, closed:     1,1000000
    **--pad** *<pad_length>*
        If set, regions in input file will be extended by pad_length bases
        on both sides.

    ###############
    Other Arguments
    ###############
    **--add-chr**
        Adds 'chr' prefix to all chromosome names in output
    **--remove-chr**
        Removes 'chr' prefix, if present, to all chromosome names in output
    

'''
import sys
import pysam
from pgpipe.genome_region import Region, RegionList, getIntervalsBetween
import pgpipe.argparse_sets
import argparse



#Given a list of CDS intervals and optional buffer length,
#Generate full set of intervals between regions.
def createParser():
    parser = argparse.ArgumentParser(description=("Generates list of "
                "intervals within a provided BED file with option "
                "to pad intervals by a fixed amount"))
    parser.add_argument('--bed', dest="region_name",
                        help="Name of gene region file")
    parser.add_argument("--bed-column-index", dest="colstr",
                        help=("Comma-separated list of length 3 with 0-based"
                        " indexes of start, end, and chromosome data in input"
                        " BED file. Default for normal BED is 1,2,0"))
    parser.add_argument('--zero-ho', dest="zeroho", action="store_true",
                        help="Region list is 1 indexed, not 0")
    parser.add_argument('--zero-closed', dest="zeroclosed", action="store_true",
                        help="Use zero-based, closed coordinates")
    parser.add_argument('--pad', dest="pad_count", default=0,
                        help="Extend input regions by provided value")
    cg = parser.add_mutually_exclusive_group()
    cg.add_argument('--addchr', dest="add_chr", action="store_true")
    cg.add_argument('--removechr',dest="remove_chr",action="store_true")
    parser.add_argument('--out', dest="output_name",
                        help="Output filename, default is stdout")
    return parser



def get_intergenic(sysargs):
    """
    Creates a BED file with regions that are not covered in the input BED.

    This function is similar to 'bedtools compliment', except it doesn't 
    require a chromosome end position file (which results in the last region 
    of a chromosome being thrown out). A flanking distance from the ends of 
    input regions can be set with --pad to truncate regions (for example, if 
    trying to find regions that are outside a 5Kb window of a gene).
    Additional options are included for dealing with zero-based, half-open
    based BED files, and for adding/removing 'chr' from chromosome names. 
    By default list will be sorted and overlapping regions will be merged.

    Parameters
    ----------
    --bed : str (required)
        Name of input BED file
    --bed-column-index : ints (3)
        Comma-separated list of length 3 with 0-based indexes of start, end,
        and chromosome data in input BED file. Default for normal BED is
        "1,2,0".
    --zero-ho : bool
        If set, treats regions in BED file as 0-based, (h)alf (o)pen 
        coordinates rather than 1-based, closed
    --pad : int
        Distance to pad input windows. If padding between two 
        consecutive regions cross, will not output region for area.
    --out : str
        Output filename

    Other Parameters
    ----------------
    --addchr : bool
        If set, will add 'chr' to all chromosome names
    --removechr : bool
        If set, will remove 'chr' from all chromosome names
        starting with it
    


    """
    parser = createParser()
    args = parser.parse_args(sysargs)
    if args.region_name is None:
        raise Exception("BED input filename required")
    reg_list = RegionList(filename=args.region_name, colstr=args.colstr,
                          zeroho=args.zeroho, zeroclosed=args.zeroclosed)
    out_list = getIntervalsBetween(reg_list, int(args.pad_count))
    out_list.printList(file_name=args.output_name, add_chr=args.add_chr)

if __name__ == "__main__":
    #initLogger()
    get_intergenic(sys.argv[1:])



#!/usr/bin/env python
'''
    Given a BED file and a VCF file, will find regions in the VCF that have
    a certain number of variant sites meeting specified criteria.

    This function checks to make sure that regions specified by other means
    have enough sites to be considered informative in either the four-gamete 
    test or an IM run. Because many variants are not considered useful in 
    these situations, filters are provided for removing sites with missing 
    data, non-biallelic sites, indels, CpGs, and singletons from determining
    if there are a sufficient number of sites in the region. Output is either
    a BED file of a set number of random regions that pass the criteria, or a 
    file with all regions that pass. It can also remove regions that are below
    a minimum specified length. If a model file is specified, only the 
    individuals in the selected population will be considered for singleton and
    missing data filters.

    ###############
    Input Arguments
    ###############
    **--vcf** *<vcf_name>*
        Name of input VCF file
    **--bed** *<bed_file>*
        Name of input BED file

    ################
    Output Arguments
    ################
    **--out** *<out_name>*
        Name of output BED file
    **--randcount** *<number_of_regions>*
        If set, will output set number of regions randomly selected
        from those that pass the criteria. Default behavior is to output
        all regions that match criteria.

    ###################
    Filtering Arguments
    ###################
    **--remove-multi**
        Do not count tri-allelic+ sites toward number of valid variants in
        a region
    **--remove-missing** *<max_misscount>*
        Do not count sites with more than max_misscount missing individuals.
        Default of -1 indicates all sites are included, 0 indicates sites
        with any missing data are not counted..
    **--remove-indels**
        Do not count indels toward number of valid variants in a region
     **--parsecpg** *<fasta_reference_filename>*
        Optional argument that if set, will detect whether or not variants
        are CpGs. A check is made to make sure the positions in the FASTA
        line up with the correct variant reference allele.   
    **--informative-count** *<minimum_allele_count>*
        Minimum number of haplotypes with both alleles at a site. Default
        is 2, meaning there must be two of each of reference and alternate
        allele in the target individuals. Can be set to 1 to filter 
        out invariant sites.
    **--minsites** *<min_sites>*
        Minimum number of variants required for a region to pass the filtering
        criteria. Variants that match specified arguments will not be counted
        towards this total.
    **--min-length** *<min_length>*
        Minimum base length of region for region to be considered.

    ################
    Region Arguments
    ################
    **--bed-column-index** *<start_idx>*,*<end_idx>*,*<chrom_idx>*
        Comma-separated string of the zero-based indices of the start, end,
        and chromosome columns in the input file, so the file doesn't need to
        be reformatted. Default for a regular BED file is 1,2,0.
    **--zero-ho**
        If set, indicates input BED regions are on zero-based, half-open
        coordinate system, as opposed to one-based, closed. For example, the
        first million bases on a chromosome would be:
            Zero-based, half-open: 0,1000000
            One-based, closed:     1,1000000
    **--pad** *<pad_length>*
        If set, regions in input file will be extended by pad_length bases
        on both sides.
    **--keep-full-line**
        If set, regions output will be the same line as was present in the 
        input file. Default behavior is to output start/end/chrom columns
        in that order, without any other data. 

    ###############
    Model Arguments
    ###############
    **--model-file** *<model_filename>*
        Name of model file that contains individuals to be considered
        for filtering. 
    **--model** *<model_name>*
        If model file contains more than one model, name of model
        to be used.

    ###############
    Other Arguments
    ###############
    **--no-sorting**
        Will output regions in order they were in input file.
        Default behavior is to sort regions before filtering.
    **--tbi** *<tabix_index>*
        If input VCF is compressed and tabix file is not default,
        provide the tabix filename here.
    **--no-xy**
        Removes regions on X/Y chromosomes from consideration.
'''

import sys
import pysam
import argparse
import os
from random import shuffle

from pgpipe.vcf_reader_func import getPassSites, VcfReader
from pgpipe.genome_region import Region, RegionList
from pgpipe.logging_module import initLogger
from pgpipe.model import Model, read_single_model

def createParser():
    parser = argparse.ArgumentParser(description=("Given a VCF file and a"
            " list of intervals (BED style), will output regions that have"
            " over a certain number of qualifying variants in the region"))
    parser.add_argument("--vcf", dest="vcfname", help=("Input VCF name"))
    parser.add_argument("--bed", dest="bedname", help=("BED filename "
                        "with regions for inspection"))
    parser.add_argument("--bed-column-index",dest="gene_col",help=("Three "
                        "comma-separated integers indicating columns for "
                        "start,end,and chromosme in BED file (default 1,2,0"))
    parser.add_argument("--out",dest="outname",help=("Output filename,"
                        "default is stdout"))
    parser.add_argument("--zero-ho", dest="zeroho", action="store_true",
                        help="BED input in zero-based, half-open format")
    parser.add_argument("--zero-closed", dest="zeroclosed",action="store_true")
    parser.add_argument("--parsecpg", dest="refname",help=("If filtering "
                        "for CpGs, provide reference filename here"))
    parser.add_argument("--remove-indels", dest="remove_indels", 
                        action="store_true", help=("Do not count indels "
                        "toward informative site count"))
    parser.add_argument("--remove-multi", dest="remove_multiallele",
                        action="store_true",help=("Do not count multiallelic"
                        "sites toward informative site count"))
    parser.add_argument("--remove-missing", dest="remove_missing", 
                        default=-1, type=int, help=("Will filter out site if"
                        " more than the given number of individuals "
                        "(not genotypes) are missing data. 0 removes sites "
                        "with any missing data, -1 (default) removes nothing"))
    parser.add_argument("--informative-count", dest="informative_count", 
                        type=int, default=2, help=("Minimum allele count of"
                        "sites counted as informative. Used so sites are "
                        "considered valid in four-gamete filtering."))
    parser.add_argument("--minsites", dest="minsites", default=3, 
                        help=("Regions with at least this many variants "
                        "passing filters will be output"))
    parser.add_argument("--tbi", dest="tabix_index", help=("Path to bgzipped "
                        "file's index if name doesn't match VCF file"))
    parser.add_argument("--randcount",dest="randcount",type=int,default=-1,
                        help=("If set, will randomly draw from input region "
                        "file until randcount # of passing BED regions "
                        "are found"))
    parser.add_argument("--no-xy",dest="filter_xy",action="store_true",
                        help="Remove X/Y chromosomes from valid regions")
    parser.add_argument("--min-length",dest="min_length",type=int,default=1000,
                        help="Minimum length of valid regions")
    parser.add_argument("--model-file",dest="modelname",help=("Model file for "
                        "selecting samples from VCF"))
    parser.add_argument("--model",dest="poptag",help="Name of pop if model "
                        "has more than one")
    parser.add_argument("--keep-full-line",dest="keep_full_line",
                        action="store_true",help=("Output BED line from "
                        "original input BED file"))
    parser.add_argument("--no-sorting",dest="sort_lists",action="store_false",
                        help=("Guarantees output regions are in same order"
                        "as input file"))
    return parser


def filter_bed_regions(sys_args):
    #parser = argparse.parse_args(sys_args)
    parser = createParser()
    args = parser.parse_args(sys_args)
    if args.outname is not None:
        outf = open(args.outname,'w')
    else:
        outf = sys.stdout

    popmodel = None
    if args.modelname is not None:
        popmodel = read_single_model(args.modelname,args.poptag)
        #popmodels = read_model_file(args.modelname)
        #if len(popmodels) != 1:
        #    popmodel = popmodels[args.poptag]
        #else:
        #    pp = list(popmodels.keys())
        #    popmodel = popmodels[pp[0]]
    
    vcf_reader = VcfReader(args.vcfname,index=args.tabix_index,popmodel=popmodel)
    fasta_seq = None
    if args.refname is not None:
        fasta_seq = pysam.FastaFile(args.refname)

    #regions = RegionList(filename=args.bedname,zeroho=args.zeroho,zeroclosed=args.zeroclosed,sortlist=(not args.randcoun))
    randomize = False
    if args.randcount != -1:
        randomize = True

    regions = RegionList(filename=args.bedname,zeroho=args.zeroho,
                         zeroclosed=args.zeroclosed,sortlist=args.sort_lists,
                         colstr=args.gene_col,
                         keep_full_line=args.keep_full_line)
    if args.filter_xy:
        regions.filterOutXY()
    remove_cpg = (True if args.refname is not None else False)
    idx_list = [i for i in range(len(regions.regions))]
    if args.randcount != -1:
        shuffle(idx_list)
    regions_output = 0
    idx_output = []
    #for region in regions.regions:
    for i in idx_list:
        region = regions.regions[i]
        if len(region) < args.min_length:
            continue
        rec_list = vcf_reader.getRecordList(region)
        pass_list = getPassSites(rec_list, remove_cpg=remove_cpg,
                    remove_indels=args.remove_indels,
                    remove_multiallele=args.remove_multiallele,
                    remove_missing=args.remove_missing,
                    inform_level=args.informative_count,
                    fasta_ref=fasta_seq)
        if pass_list.count(True) >= int(args.minsites):
            #print (region.toStr(sep='\t'))
            idx_output.append(i)
            regions_output += 1
        if args.randcount != -1 and regions_output == args.randcount:
            break

    if args.randcount != -1 and regions_output != args.randcount:
        sys.stderr.write("Only %d of %d regions found\n"%(regions_output,args.randcount))
        #exit(1)

    if randomize:
        idx_output.sort()
    
    if regions.header is not None:
        outf.write(regions.header+'\n')
    for i in idx_output:
        outf.write(regions.regions[i].toStr(sep='\t')+'\n')


if __name__ == '__main__':
    initLogger()
    filter_bed_regions(sys.argv[1:])



#!/usr/bin/env python
'''
    Automates the sampling process of a specified statistic output file. The
    function allows the user to select both the statistic in question and
    the sampling scheme. Please note that all sampling is done without
    replacement.

    ##################
    Command-line Usage
    ##################
    The statistic sampler may be called using the following command:

    .. code-block:: bash
        
        python stat_sampler.py

    *************
    Example usage
    *************
    Randomly sampling 20 windows from a windowed Fst statistic file 
    **merged_chr1_10000.windowed.weir.fst**.

    .. code-block:: bash
        
        python admixture.py --statistic-file merged_chr1_10000.windowed.weir.fst 
        --calc-statistic windowed-weir-fst --sampling-scheme random --sample-size 20

    ############################
    Input Command-line Arguments
    ############################
    **--statistic-file** *<statistic_filename>*
        Argument used to define the filename of the statistic file for sampling.

    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**.
        Cannot be used if multiple output files are created.
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--overwrite**
        Argument used to define if previous output should be overwritten.

    ###############################
    Sampling Command-line Arguments
    ###############################
    **--calc-statistic** *<windowed-weir-fst, TajimaD, window-pi>*
        Argument used to define the statistic to be sampled. Windowed Fst 
        (windowed-weir-fst), Tajima's D (TajimaD), and windowed nucleotide 
        diversity (window-pi).
    **--sampling-scheme** *<random, uniform>*
        Argument used to define the sampling scheme. Random [Default]
        sampling or uniform sampling across of number of equal-sized
        bins.
    **--uniform-bins** *<bin_int>*
        Argument used to define the number of bins in uniform sampling.
    **--sample-size** *<sample_size_int>*
        Argument used to define the total sample size. If using the uniform 
        sampling scheme, this number must be divisible by the number of bins.
    **--random-seed** *<seed_int>*
        Argument used to define the seed value for the random number generator.
'''

import os
import sys
import argparse
import itertools
import logging
import shutil
import copy
import numpy as np
import pandas as pd
from collections import defaultdict

# Insert Jared's directory path, required for calling Jared's functions. Change when directory structure changes.
#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

# Import log initializer
from pgpipe.logging_module import initLogger, logArgs

# Import basic vcftools functions
from pgpipe.vcftools import *

# Import basic vcf
from pgpipe.bcftools import check_for_index, create_index

# Model file related functions
from pgpipe.model import read_model_file

def sampler_parser(passed_arguments):
    '''
    Stat Sampler Phase Argument Parser

    Assign the parameters for stat sampler using argparse.

    Parameters
    ----------
    passed_arguments : list, optional
        Parameters passed by another function. sys.argv is used if
        not given. 

    Raises
    ------
    IOError
        If the input, or other specified files do not exist
    '''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string = None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    sampler_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Other input arguments.
    sampler_parser.add_argument('--statistic-file', help='Defines the filename of the statistic file for sampling', required = True, type = str, action = parser_confirm_file())

    # Statistic based arguments.
    statistic_list = ['windowed-weir-fst', 'TajimaD', 'window-pi']
    sampler_parser.add_argument('--calc-statistic', metavar = metavar_list(statistic_list), help = 'Defines the statistic to be sampled', type=str, choices = statistic_list)

    # Sampling methods. Currently mutually exclusive to only allow a single sampling method
    sampling_list = ['uniform', 'random']
    sampling_default = 'random'
    sampler_parser.add_argument('--sampling-scheme', metavar = metavar_list(sampling_list), help = 'Defines the sampling scheme', type=str, choices = sampling_list, default = sampling_default)

    # Sampling options
    sampler_parser.add_argument('--uniform-bins', help="Defines the number of bins in uniform sampling", type = int)
    sampler_parser.add_argument('--sample-size', help="Defines the total sample size. If using the uniform sampling, this number must be divisible by the bins.", type = int)

    # Random options
    sampler_parser.add_argument('--random-seed', help="Defines the seed value for the random number generator", type = int, default = np.random.randint(1, high = 1000000000))

    # Output arguents
    sampler_parser.add_argument('--out', help = 'Defines the complete output filename, overrides **--out-prefix**', type = str)

    sampler_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', type = str, default = 'out')

    # General arguments
    sampler_parser.add_argument('--overwrite', help = "Defines if previous output should be overwritten", action = 'store_true')


    if passed_arguments:
        return sampler_parser.parse_args(passed_arguments)
    else:
        return sampler_parser.parse_args()

def random_vcftools_sampler (stat_file_data, sample_size, with_replacements = False):
    '''
        Random Sampler.

        Returns a list of randomly selected elements from stat_file_data. The
        length of this list is defined by sample_size.

        Parameters
        ----------
        stat_file_data : list
            List of statistic values
        sample_size : int
            Size of the total sample
        with_replacements : bool, optional
            If the function should sample with replacements

        Returns
        -------
        random_samples : list
            List of the randomly sampled statistics

    '''

    # Use numpy choice to randomly sample the stat_file_data
    random_samples = np.random.choice(stat_file_data, int(sample_size), replace = with_replacements)

    # Convert the samples to a list
    random_samples = list(random_samples)

    # Return the sample list
    return random_samples

def uniform_vcftools_sampler (stat_file_data, uniform_bins, sample_size):
    '''
        Uniform Sampler.

        Seperates the elements of stat_file_data into equal sized bins (number
        of bins specified by uniform_bins arg). The samples in the bins are
        then randomly selected. The number of selected samples is defined by
        sample_size / uniform_bins.

        Parameters
        ----------
        stat_file_data : list
            List of statistic values
        uniform_bins : int
            Number of uniform bins
        sample_size : int
            Size of the total sample

        Returns
        -------
        uniform_samples : list
            List of the uniformly sampled statistics
    '''

    # Create defaultdict to store the binned data
    binned_sample_database = defaultdict(list)

    # Bin the data and get the sizes of the uniform bin catagories
    binned_samples, bin_catagories = pd.cut(stat_file_data, uniform_bins, labels = False, retbins = True)

    # Loop the bins for logging purposes
    for bin_number, bin_edge_pos in enumerate(range(uniform_bins)):
        # Store the right edge of the bin
        right_edge = bin_catagories[bin_edge_pos]
        # Store the left edge of the bin
        left_edge = bin_catagories[bin_edge_pos + 1]
        # Log the bins for the user
        logging.info('Bin %s: %s < statistic <= %s' % (bin_number + 1, right_edge, left_edge))

    # Get the number of samples to sample per bin
    samples_per_bin = sample_size / uniform_bins

    # Loop the binned data
    for vcftools_sample, sample_bin in zip(range(len(stat_file_data)), binned_samples):
        # Populate the defaultdict with the bins
        binned_sample_database[sample_bin].append(vcftools_sample)

    # Create a list to hold the sampled data
    uniform_samples = []

    # Loop the defaultdict
    for sample_bin, current_samples in binned_sample_database.items():

        # Check if the current bin needs to be sampled
        if len(current_samples) >= samples_per_bin:

            # Randomly sample the bin
            bin_samples = random_vcftools_sampler(current_samples, samples_per_bin)

        # Check if the bin is too small for sampling
        else:
            # Return the current_samples
            bin_samples = current_samples

        uniform_samples.extend(bin_samples)

    return uniform_samples

def assign_statistic_column (sample_headers, statistic):
    '''
        Assigns the statistic column.

        Determines the column with the associated values for the specified
        statistic. This function is required only for the uniform sampler, as
        the random sampler does not need to know the values to function. Due to
        the nature of the statistic files, a conversion system is required to
        determine the associated column header for each specified statistic.
        This is done using the statistic_converter variable, which will need to
        be updated along with --calc-statistic to add more statistics.

        Parameters
        ----------
        sample_headers : list
            List of headers from the statistic file
        statistic : str
            The specified statistic

        Returns
        -------
        uniform_samples : list
            List of the uniformly sampled statistics

        Raises
        ------
        Exception
            Statistic not found. Converter likely needs to be updated
        Exception
            Specified statistic not found in headers. Likely wrong statistic
            was selected
    '''

    # Defaultdict to hold the statistic conversion data.
    # Conversion: --calc-statistic value : header value
    statistic_converter = {'windowed-weir-fst':'MEAN_FST',
                           'TajimaD':'TajimaD',
                           'window-pi':'PI'}

    # Return error if the statistic is not found
    if statistic not in statistic_converter:
        raise Exception('Statistic not found. Statistic list needs to be '
                        'updated. Please contact the PPP Team.')

    # Return error if the wrong statistic was selected
    if statistic_converter[statistic] not in sample_headers:
        raise ValueError('Statistic selected not found in file specified by '
                         '--statistic-file.')

    # Save the converted statistic
    converted_statistic = statistic_converter[statistic]

    # Return the converted statistic
    return converted_statistic

def run (passed_arguments = []):
    '''
    Statistics sampler.

    Automates the sampling process of a specified statistic output file. The
    function allows the user to select both the statistic in question and
    the sampling scheme. Please note that all sampling is done without
    replacement.

    Parameters
    ----------
    --statistic-file : str
        Specifies the statistic file for filtering
    --calc-statistic : str
        Specifies the statistic to calculate. Choices: windowed-weir-fst
        (default) and TajimaD
    --statistic-window-size : int
        Specifies the window size of the statistic if not specified in the
        file
    --sampling-scheme : str
        Specifies the sampling scheme to use. Choices: random (default) and
        uniform
    --uniform-bins : int
        Specifies the number of bins for the uniform sampler
    --sample-size : int
        Specifies the total sample size. Note: If using the uniform sampling
        scheme, this number must be divisible by number of uniform bins
    --random-seed : int
        Specifies the random seed value for the random number generator
    --out : str
        Specifies the sampled statistic output filename
    --out-prefix : str
        Specifies the sampled statistic output filename prefix

    Raises
    ------
    IOError
        Output file already exists
    ValueError
        Sample size larger than the number of samples
    ValueError
        Sample size not divisible by the bin count
    ValueError
        Window size argument conflicts with values
    TypeError
        Window size argument not defined (if necessary)
    '''

    # Get arguments from command line
    sampler_args = sampler_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(sampler_args, func_name = 'stat_sampler')

    # Set the random seed
    np.random.seed(sampler_args.random_seed)

    # Check if a sample size has been specified
    if not sampler_args.sample_size:
        raise Exception('No sample size specified. Please use --sample-size')

    # Set the output filename using the out-prefix
    sampled_out_filename = sampler_args.out_prefix + '.sampled'

    # Check if the user has defined a specifed output filename
    if sampler_args.out:
        # Rename the output
        sampled_out_filename = sampler_args.out

    # Check if user has selected to overwrite previous output
    if not sampler_args.overwrite:

        # Check if output file already exists
        if os.path.isfile(sampled_out_filename):
            raise IOError('Sample file already exists')

    # Read in the sample file
    stat_file_data = pd.read_csv(sampler_args.statistic_file, sep = '\t')

    # Confirm statistic is correct, and assign the column ID
    assigned_statistic = assign_statistic_column(list(stat_file_data), sampler_args.calc_statistic)

    # Remove NaN datapoints for the statistic
    stat_file_data = stat_file_data.dropna(subset=[assigned_statistic])

    logging.info('Sample (i.e. statistic) file assigned')

    # Confirm there are enough samples
    if len(stat_file_data) < sampler_args.sample_size:
        raise ValueError('Sample size larger than the number of datapoints '
                         'within statistic file')

    # UPDATE Planned
    # Improve from equal to too few samples (i.e. samples = 100, sample_size = 95)
    elif len(stat_file_data) == sampler_args.sample_size:

        # Warns the user of poor sampling
        logging.warning('Sample size equal to number of datapoints within '
                        'statistic file')

    # Run the uniform sampler
    if sampler_args.sampling_scheme == 'uniform':

        # Check that uniform bins have been specified
        if not sampler_args.uniform_bins:
            raise Exception('No uniform bin count specified. Please use --uniform-bin')

        # Check that the sample size is divisible by the bins
        if sampler_args.sample_size % sampler_args.uniform_bins != 0:
            raise ValueError('Sample size not divisible by the bin count')

        selected_samples = sorted(uniform_vcftools_sampler(list(stat_file_data[assigned_statistic]), sampler_args.uniform_bins, sampler_args.sample_size))

        if len(selected_samples) < sampler_args.sample_size:
            logging.warning('Uniform sampler was unable to reach sample size. '
                            'Likely too few samples per bin, consider reducing '
                            'the number of bins or the sample size.')

        logging.info('Uniform sampling complete')

    # Run the random sampler
    if sampler_args.sampling_scheme == 'random':

        selected_samples = sorted(random_vcftools_sampler(list(stat_file_data.index), sampler_args.sample_size))

        logging.info('Random sampling complete')

    # Reduce to only selected samples
    sampled_samples = stat_file_data[stat_file_data.index.isin(selected_samples)].copy()

    # Create selected samples TSV file with the defined output filename
    sampled_samples.to_csv(sampled_out_filename, sep = '\t', float_format = '%g', index = False)

    logging.info('Created selected samples file')

if __name__ == "__main__":
    initLogger()
    run()



import os
import sys
import subprocess
import shutil
import argparse
import glob
import copy
import logging
import re

# Call PPP-based scripts
#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir,'pppipe')))

from pgpipe.vcf_reader_func import checkFormat
from pgpipe.bcftools import convert_vcf
from pgpipe.logging_module import initLogger, logArgs

def log_plink_reference (out_filename, append_mode = False, ref_header = True):

    # Check if the file is to be written in append mode
    if append_mode:

        # Open the file
        log_file = open(out_filename + '.log', 'a')

        # Check if the ref header should be added
        if ref_header:
            log_file.write('\nPlease Reference alongside the PPP:\n')

    else:

        # Open the file
        log_file = open(out_filename + '.log', 'w')

        # Check if the ref header should be added
        if ref_header:
            log_file.write('Please Reference alongside the PPP:\n')

    log_file.write('Chang, C. C. et al. Second-generation PLINK: Rising to ' 
                   'the challenge of larger and richer datasets. Gigascience ' 
                   '(2015). doi:10.1186/s13742-015-0047-8')
    log_file.close()

    logging.info('Reference assigned')

def assign_delim_from_ids (filename, id_column = 0, default_delim = '_'):

    delim_symbols = ['-', '|', '/', '@', '&', '*', '^', '%', '$']

    symbols_present = []

    with open(filename, 'rU') as sample_file:
        for line_pos, sample_line in enumerate(sample_file):
            if line_pos > 1:
                # Stores the current sample ID
                sample_id = sample_line.strip().split()[id_column]
                # Symbols found in the sample ID
                id_symbols = re.sub(r'[a-zA-Z0-9]','', sample_id)
                # Iterate the symbols from the ID
                for id_symbol in iter(id_symbols):
                    # Check if that symbol has been stored
                    if id_symbol not in symbols_present:
                        # Store symbols
                        symbols_present.append(id_symbol)

    if default_delim not in symbols_present:
        return default_delim
    else:
        for delim_symbol in delim_symbols:
            if delim_symbol not in symbols_present:
                return delim_symbol

    raise IOError('Cannot assign plink delimiter')

def assign_plink_output_args (out_prefix, out_format, overwrite = True):
    '''
        Assigns the output arguments

        Parameters
        ----------
        filename : str
            Specifies the input filename of unknown format

        Returns
        -------
        list
            Returns vcftools input command for `filename`

        Raises
        ------
        IOError
            If filename is an unknown file format
    '''
    # Check if previous output should be overwritten
    if not overwrite:

        # Check if the output is only a single file
        if out_format in ['vcf', 'vcf.gz', 'bcf']:

            # Save the filename of the expected output
            out_filename = out_prefix + '.' + out_format

            # Check if intermediate already exists
            if out_format == 'bcf':

                # Save the filename of the expected intermediate
                intermediate_filename = out_prefix + '.vcf.gz'
                raise Exception('%s intermediate exists. Add --overwrite to ignore' % intermediate_filename)

            # Check if the output already exists
            if os.path.isfile(out_filename):
                raise Exception('%s already exists. Add --overwrite to ignore' % out_filename)

        # Check if the output if a ped or bed file-set
        elif out_format in ['ped', 'ped-12', 'binary-ped']:

            if out_format == 'ped' or out_format == 'ped-12':

                # List of the output suffixes for ped files
                out_suffixes = ['ped', 'map']

            elif out_format == 'binary-ped':

                # List of the output suffixes for bed files
                out_suffixes = ['bed', 'bim', 'fam']

            # Loop the suffixes
            for out_suffix in out_suffixes:

                # Save the filename of the expected output
                out_filename = out_prefix + '.' + out_suffix

                # Check if the output already exists
                if os.path.isfile(out_filename):
                    raise Exception('%s already exists. Add --overwrite to ignore' % out_filename)

    # Check if the output is only a single file
    if out_format in ['vcf', 'vcf.gz', 'bcf']:

        if out_format == 'vcf':
            return ['--recode', 'vcf-iid', '--out', out_prefix]

        if out_format == 'vcf.gz':
            return ['--recode', 'vcf-iid', 'bgz', '--out', out_prefix]

        if out_format == 'bcf':
            return ['--recode', 'vcf-iid', 'bgz', '--out', out_prefix]

    # Check if the output if a ped or bed file-set
    elif out_format in ['ped', 'ped-12', 'binary-ped']:

        if out_format == 'ped':
            return ['--recode', '--out', out_prefix]

        if out_format == 'ped-12':
            return ['--recode12', '--out', out_prefix]

        if out_format == 'binary-ped':
            return ['--make-bed', '--out', out_prefix]

    else:
        raise IOError('Unknown file format. This error should NOT show up')

def confirm_ped_files (ped_filename, map_filename):

    # Check if a bed file is assigned
    if not ped_filename:
        raise IOError('Unable to assign ped file. Please confirm the file is named correctly')

    # Check if bim and fam file are not assigned
    if not map_filename:
        raise IOError('Unable to assign map file. Please confirm the map file (i.e. --map) is assigned')

    # Return True if all checks passed
    return True

def confirm_ped_prefix (ped_prefix):

    # Check if the a prefix was assigned
    if not ped_prefix:
        raise IOError('No PED prefix specified. Please confirm the prefix is specified correctly')
    
    # Determine the input files from the prefix
    ped_prefix_filenames = glob.glob(ped_prefix + '*')

    # Check if files were identified, if not return error message
    if not ped_prefix_filenames:
        raise IOError('Unable to assign input files from PED prefix. Please confirm the prefix is specified correctly')

    # Empty filenames, required if not found
    ped_filename = None
    map_filename = None
    
    # Check if expected files are found
    for ped_prefix_filename in ped_prefix_filenames:

        # Check for PED and MAP files
        if ped_prefix_filename.endswith('.ped'):
            ped_filename = ped_prefix_filename
        elif ped_prefix_filename.endswith('.map'):
            map_filename = ped_prefix_filename

    # Check that input files included a ped file
    if not ped_filename:
        # Return error message if no ped input was found
        raise IOError('Unable to assign ped file. Please confirm the prefix and/or command is specified correctly')

    if not map_filename:
        # Return error message if no map input was found
        raise IOError('Unable to assign map file. Please confirm the file is named correctly')

    # Return True if all checks passed
    return True

def confirm_bed_files (bed_filename, bim_filename, fam_filename):

    # Check if a bed file is assigned
    if not bed_filename:
        raise IOError('Unable to assign binary-ped file. Please confirm the file is named correctly')

    # Check if the bim and fam files are not assigned
    if not bim_filename and not fam_filename:
        raise IOError('Unable to assign the bim and fam files. Please confirm the files (i.e. --bim, --fam) are assigned')

    # Check if the bim is not assigned
    if not bim_filename:
        raise IOError('Unable to assign bim file. Please confirm the bim file (i.e. --bim) is assigned')

    # Check if the fam file is not assigned
    if not fam_filename:
        raise IOError('Unable to assign fam file. Please confirm the fam file (i.e. --fam) is assigned')

    # Return True if all the checks passed
    return True

def confirm_bed_prefix (bed_prefix):

    # Check if the a prefix was assigned
    if not bed_prefix:
        raise IOError('No Binary-PED prefix specified. Please confirm the prefix is specified correctly')
    
    # Determine the input files from the prefix
    bed_prefix_filenames = glob.glob(bed_prefix + '*')

    # Check if files were identified, if not return error message
    if not bed_prefix_filenames:
        raise IOError('Unable to assign input files from Binary-PED prefix. Please confirm the prefix is specified correctly')

    # Empty filenames, required if not found
    bed_filename = None
    bim_filename = None
    fam_filename = None
    
    # Check if expected files are found
    for bed_prefix_filename in bed_prefix_filenames:

        # Check for BED and associated files
        if bed_prefix_filename.endswith('.bed'):
            bed_filename = bed_prefix_filename
        elif bed_prefix_filename.endswith('.bim'):
            bim_filename = bed_prefix_filename
        elif bed_prefix_filename.endswith('.fam'):
            fam_filename = bed_prefix_filename

    # Check that input files included a bed file
    if not bed_filename:
        # Return error message if no ped/bed input was found
        raise IOError('Unable to assign bed file. Please confirm the prefix and/or command is specified correctly')

    # Check that input files included a bed file
    if not bim_filename:
        # Return error message if no bim input was found
        raise IOError('Unable to assign bim file. Please confirm the file is named correctly')

    # Check that input files included a bed file
    if not fam_filename:
        # Return error message if no fam input was found
        raise IOError('Unable to assign fam file. Please confirm the file is named correctly')

    # Return True if all checks passed
    return True
    
def convert_vcf_to_plink (vcf_filename, vcf_fid, out_prefix, out_format, threads = 1, overwrite = False):

    logging.info('Beginning VCF conversion')

    # List of arguments for bed converions
    convert_vcf_args = []

    # Check if a specific number of threads have been assigned
    if threads:

        convert_vcf_args.extend(['--threads', threads])

    # Assign the vcf file format
    vcfname_format = checkFormat(vcf_filename)

    # Check if a family ID was assigned for the vcf file
    if vcf_fid:

        # Assign the family id
        convert_vcf_args.extend(['--const-fid', vcf_fid])

    # If no family ID was assigned, use double IDs
    else:

        convert_vcf_args.append('--double-id')

    convert_vcf_args.append('--allow-extra-chr')

    # Assign the associated input command, or return an error.
    if vcfname_format == 'vcf' or vcfname_format == 'bgzip':

        # Add the bed associated commands
        convert_vcf_args.extend(['--vcf', vcf_filename])

    elif vcfname_format == 'bcf':

        # Add the bed associated commands
        convert_vcf_args.extend(['--bcf', vcf_filename])

    else:
        raise Exception('Unknown VCF file format')

    # Assign the output arguments
    vcf_output_args = assign_plink_output_args(out_prefix, out_format, overwrite)

    # Add the output args to the ped call
    convert_vcf_args.extend(vcf_output_args)

    # Call plink with the selected arguments
    call_plink(convert_vcf_args, out_prefix, out_format)

    logging.info('Finished conversion')

def convert_ped (ped_filename = None, map_filename = None, ped_prefix = None, out_prefix = None, out_format = None, threads = 1, overwrite = False, **kwargs):
    
    logging.info('Beginning PED conversion')

    # List of arguments for bed converions
    convert_ped_args = []

    # Check if a specific number of threads have been assigned
    if threads:

        convert_ped_args.extend(['--threads', threads])

    # Create blank list to hold ped input arguments
    ped_input_args = []

    # Check if the a prefix was assigned
    if ped_prefix and confirm_ped_prefix(ped_prefix):
    
        # Assign bed input arguments from a prefix
        ped_input_args = ['--file', ped_prefix]

    # Check if a bed file is assigned
    elif ped_filename and confirm_ped_files(ped_filename, map_filename):

        # Assign bed input arguments from files
        ped_input_args = ['--ped', ped_filename, '--map', map_filename]

    # Confirm input arguments were assigned
    if not ped_input_args:
        raise Exception('Unable to assign PED input arguments. Please confirm the arguments are specified correctly')

    # Add the ped input arguments
    convert_ped_args.extend(ped_input_args)

    # Assign the output arguments
    ped_output_args = assign_plink_output_args(out_prefix, out_format, overwrite)

    # Add the output args to the ped call
    convert_ped_args.extend(ped_output_args)

    # Call plink with the selected arguments
    call_plink(convert_ped_args, out_prefix, out_format)

    logging.info('Finished conversion')

def convert_bed (bed_filename = None, bim_filename = None, fam_filename = None, bed_prefix = None, out_prefix = None, out_format = None, threads = 1, overwrite = False, **kwargs):
        
    logging.info('Beginning Binary-PED (i.e. BED) conversion')

    # List of arguments for bed converions
    convert_bed_args = []

    # Check if a specific number of threads have been assigned
    if threads:

        convert_bed_args.extend(['--threads', threads])

    # Create blank list to hold bed input arguments
    bed_input_args = []

    # Check if the a prefix was assigned
    if bed_prefix and confirm_bed_prefix(bed_prefix):
    
        # Assign bed input arguments from a prefix
        bed_input_args =  ['--bfile', bed_prefix]

    # Check if a bed file is assigned
    elif bed_filename and confirm_bed_files(bed_filename, bim_filename, fam_filename):

        # Assign bed input arguments from files
        bed_input_args = ['--bed', bed_filename, '--bim', bim_filename, '--fam', fam_filename]

    # Confirm input arguments were assigned
    if not bed_input_args:
        raise Exception('Unable to assign Binary-PED input arguments. Please confirm the arguments are specified correctly')

    # Add the bed input arguments
    convert_bed_args.extend(bed_input_args)

    # Assign the output arguments
    bed_output_args = assign_plink_output_args(out_prefix, out_format, overwrite)

    # Add the output args to the ped call
    convert_bed_args.extend(bed_output_args)

    # Call plink with the selected arguments
    call_plink(convert_bed_args, out_prefix, out_format)

    logging.info('Finished conversion')

def convert_haps_to_vcf (haps_prefix, output_format, output_prefix = ''):

    # Check that input files included a haps file
    if not os.path.isfile(haps_prefix + '.haps'):
        # Return error message if no haps input was found
        raise IOError('Unable to assign haps file. Please confirm the prefix and/or command is specified correctly')

    # Check that input files included a sample file
    if not os.path.isfile(haps_prefix + '.sample'):
        # Return error message if no sample input was found
        raise IOError('Unable to assign sample file. Please confirm the prefix and/or command is specified correctly')

    haps_input_args = ['--haps', haps_prefix + '.haps', 'ref-first', '--sample', haps_prefix + '.sample']

    # If no output_prefix is assigned, use haps_prefix
    if not output_prefix:
        output_prefix = haps_prefix

    plink_delim = assign_delim_from_ids(haps_prefix + '.sample')

    # Assign general command line arguments
    haps_output_args = ['--recode', 'vcf-iid', 'id-delim=' + plink_delim, '--out', output_prefix]

    # Check if a compressed file should be created
    if output_format in ['vcf.gz', 'bcf']:

        # Insert the compression argument
        haps_output_args.insert(2, 'bgz')

    # Call plink2
    standard_plink2_call(haps_input_args + haps_output_args)

    # Convert VCFGZ to BCF, as PLINK cannot directly create a BCF
    if output_format == 'bcf':

        # Convert to BCF
        convert_vcf(output_prefix + '.vcf.gz', output_prefix, output_format, overwrite = True)

def check_plink_for_errors (plink_stderr):
    '''
        Checks the plink stderr for errors

        Parameters
        ----------
        plink_stderr : str
            plink stderr

        Raises
        ------
        IOError
            If plink stderr returns an error
    '''

    # Print warning, if found
    if 'Warning' in plink_stderr:
        logging.warning(plink_stderr)

    # Print output if error found. Build up as errors are discovered
    elif plink_stderr:
        raise Exception(plink_stderr)

def standard_plink2_call (plink2_call_args):
    '''
        Standard call of plink2

        The function calls plink2. Returns the stderr of plink to create a log
        file of the call.

        Parameters
        ----------
        plink2_call_args : list
            plink2 arguments

        Raises
        ------
        Exception
            If plink2 stderr returns an error
    '''


    # plink subprocess call
    plink2_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),'bin','plink2')
    plink2_call = subprocess.Popen([plink2_path] + list(map(str, plink2_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Wait for plink to finish
    plink2_out, plink2_err = plink2_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        plink2_out = plink2_out.decode()
        plink2_err = plink2_err.decode()

    logging.info('plink2 call complete')

    # Check that the log file was created correctly
    check_plink_for_errors(plink2_err)

def standard_plink_call (plink_call_args):
    '''
        Standard call of plink

        The function calls plink. Returns the stderr of plink to create a log
        file of the call.

        Parameters
        ----------
        plink_call_args : list
            plink arguments

        Raises
        ------
        Exception
            If plink stderr returns an error
    '''

    # Create a string with the plink path

    plink_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'bin', 'plink')

    # plink subprocess call
    plink_call = subprocess.Popen([plink_path] + list(map(str, plink_call_args)), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # Wait for plink to finish
    plink_out, plink_err = plink_call.communicate()

    # Check if code is running in python 3
    if sys.version_info[0] == 3:
        # Convert bytes to string
        plink_out = plink_out.decode()
        plink_err = plink_err.decode()

    logging.info('plink call complete')

    # Check that the log file was created correctly
    check_plink_for_errors(plink_err)

def call_plink (plink_call_args, output_prefix = None, output_format = None):
    '''
        Determines which version of plink (plink1.9 vs. plink2a) to call

        The function is required for the plink wrapper to determine the needed
        plink version and also automates conversion to BCF, which both versions
        of plink do not support.

        Parameters
        ----------
        plink_call_args : list
            plink arguments
        output_prefix : str
            Output prefix for non-plink conversions
        output_format : str
            Output format for non-plink conversions

    '''

    # Check if the call requries plink2 to operate. This function should be
    # removed once plink2 is out of alpha
    if '--haps' in plink_call_args:

        # Convert vcf id system for plink 1.9 to plink 2
        if 'vcf-iid' in plink_call_args:
            format_index = plink_call_args.index('vcf-iid')
            plink_call_args[format_index:format_index + 1] = ['vcf', 'id-paste=fid']

        # Call the plink2 subprocess function
        standard_plink2_call(plink_call_args)

    # Call plink1.9
    else:

        # Call the plink subprocess function
        standard_plink_call(plink_call_args)

    # Convert VCFGZ to BCF, as PLINK cannot directly create a BCF
    if output_format == 'bcf':

        # Convert to BCF
        convert_vcf(output_prefix + '.vcf.gz', output_prefix, output_format)


import sys
import argparse
import pysam
import gzip

from pgpipe.vcf_reader_func import VcfReader
from pgpipe.model import Model, read_single_model

def createParser():
    parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
    parser.add_argument("--vcfs",nargs="+",help=("VCFs that represent loci for input"))
    parser.add_argument("--model-file",dest="model_file",help="Name of model file")
    parser.add_argument("--model",dest="model",help="Name of model in model file if more than one model is present")
    parser.add_argument("--out",dest="outname",help="Name of output treemix file")
    parser.add_argument("--pad",dest="pad",type=int,default=200,help=("Loci will be padded to this many sites (must be larger than locus with most variant sites"))
    parser.add_argument("--compress-out",dest="compress",action="store_true",help=("Will compress output, adding .gz to output name if not already specified"))

    return parser

def createVarString(rec,vcf_reader):
    variant_site = False
    outstr = ''
    for p in vcf_reader.popkeys.keys():
        refcount = 0
        altcount = 0
        for indiv_idx in vcf_reader.popkeys[p]:
            for hap in range(len(rec.samples[indiv_idx].alleles)):
                if rec.samples[indiv_idx].alleles[hap] != rec.ref:
                    altcount += 1
                    variant_site = True
                else:
                    refcount += 1
        outstr += str(refcount)+','+str(altcount)+' '
    return outstr.strip(),variant_site

def createInvarString(rec,vcf_reader):
    outstr = ''
    for p in vcf_reader.popkeys.keys():
        refcount = 0
        for indiv_idx in vcf_reader.popkeys[p]:
            refcount += len(rec.samples[indiv_idx].alleles)
        outstr += str(refcount)+',0 '
    return outstr.strip()

def createPopString(vcf_reader):
    return ' '.join(vcf_reader.popkeys.keys())

def writeToFile(outf,line,compress):
    if compress:
        line = line.encode()
    outf.write(line)

def vcf_treemix_convert(sysargs):
    parser = createParser()
    args = parser.parse_args(sysargs)
    popmodel = read_single_model(args.model_file,args.model)
    if args.compress:
        outfn = (args.out+'.gz' if args.out[-3:] != '.gz' else args.out)
        output_file = gzip.open(outfn,'wb')
    else:
        output_file = open(args.outname,'w')

    pop_out = False

    for vcfname in args.vcfs:
        vcf_reader = VcfReader(vcfname,popmodel=popmodel)
        rec_list = vcf_reader.getRecordList()
        vcf_reader.reader.close()
        var_count = 0
        if not pop_out:
            writeToFile(output_file,createPopString(vcf_reader)+'\n',args.compress)
            pop_out = True
        for rec in rec_list:
            sitestr,varsite = createVarString(rec,vcf_reader)
            if varsite:
                writeToFile(output_file,sitestr+'\n',args.compress)
                var_count += 1
        if var_count > args.pad:
            raise Exception("%s has %d loci, over max of %d"%(vcfname,var_count,args.pad))
        for i in range(var_count,args.pad):
            writeToFile(output_file,createInvarString(rec_list[0],vcf_reader)+'\n',args.compress)







if __name__ == "__main__":
    vcf_treemix_convert(sys.argv[1:])


#!/usr/bin/env python
'''
    Automates VCF file filering using BCFtools. Filters may be used independently or
    combined to create complex filtering operations.        

    ##################
    Command-line Usage
    ##################
    The VCF file filter may be called using the following command:

    .. code-block:: bash
        
        python vcf_filter.py

    *************
    Example usage
    *************
    Command-line to only include biallelic sites:

    .. code-block:: bash
        
        python vcf_filter.py --vcf input.bcf --filter-min-alleles 2 --filter-max-alleles 2

    Command-line to only include chromosome chr1:

    .. code-block:: bash
        
        python vcf_filter.py --vcf input.vcf.gz --filter-include-chr chr1

    Command-line to remove indels and ouput a BCF file:

    .. code-block:: bash
        
        python vcf_filter.py --vcf input.vcf --filter-exclude-indels --out-format bcf

    ############
    Dependencies 
    ############
    * `BCFtools <https://samtools.github.io/bcftools/bcftools.html>`_    

    ############################
    Input Command-line Arguments
    ############################
    **--vcf** *<input_filename>*
        Argument used to define the filename of the VCF file to be filtered.
    **--model-file** *<model_filename>*
    	Argument used to define the model file. Please note that this argument cannot be 
    	used with the individual-based filters.
    **--model** *<model_str>*
    	Argument used to define the model (i.e. the individual(s) to include). Please note
    	that this argument cannot be used with the individual-based filters.

    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--out-format** *<vcf, vcf.gz, bcf, bed, sites>*
        Argument used to define the desired output format. Formats include: uncompressed 
        VCF (vcf); compressed VCF (vcf.gz) [default]; BCF (bcf); variants in bed format; 
        or variants in sites format.
    **--overwrite**
        Argument used to define if previous output should be overwritten.

    #############################
    Filter Command-line Arguments
    #############################
    The filtering arguments below are roughly seperated into catagoires. Please not that
    mulitple filters are seperated into two opposing function types **include** and 
    **exclude**.

    **************************
    Individual-Based Arguments
    **************************
    Please note that all individual-based arguments are not compatible with either the 
    **--model** or **--model-file** command-line arguments.

    **--filter-include-indv** *<indv_str>* *<indv1_str, indv2_str, etc.>*
        Argument used to define the individual(s) to include. This argument may be used 
        multiple times if desired.
    **--filter-exclude-indv** *<indv_str>* *<indv1_str, indv2_str, etc.>*
        Argument used to define the individual(s) to exclude. This argument may be used 
        multiple times if desired.
    **--filter-include-indv-file** *<indv_filename>*
        Argument used to define a file of individuals to include.
    **--filter-exclude-indv-file** *<indv_filename>*
        Argument used to define a file of individuals to exclude.

    *******************************
    Allele/Genotype-Based Arguments
    *******************************
    **--filter-min-alleles** *min_int*
    	Argument used to include positions with a number of allele >= to the 
    	given number.
    **--filter-max-alleles** *max_int*
        Argument used to include positions with a number of allele <= to the 
        given number.
    **--filter-maf-min** *maf_proportion*
    	Argument used to include sites with equal or greater MAF values.
    **--filter-maf-max** *maf_proportion*
    	Argument used to include sites with equal or lesser MAF values.
    **--filter-mac-min** *mac_int*
    	Argument used to include sites with equal or greater MAC values.
    **--filter-mac-max** *mac_int*
    	Argument used to include sites with equal or lesser MAC values.
    **--filter-include-indels**
    	Argument used to include positions if they contain an insertion or a deletion. 
    **--filter-exclude-indels**
    	Argument used to exclude positions if they contain an insertion or a deletion.
    **--filter-include-snps**
        Argument used to include positions if they contain a SNP. 
    **--filter-exclude-snps**
        Argument used to exclude positions if they contain a SNP. 
    **--filter-include-snp** *<rs#>* *<rs#1, rs#2, etc.>*
    	Argument used to include SNP(s) with the matching ID. This argument may be used 
        multiple times if desired.
    **--filter-exclude-snp** *<rs#>* *<rs#1, rs#2, etc.>*
        Argument used to exclude SNP(s) with the matching ID. This argument may be used 
        multiple times if desired.
    **--filter-include-snp-file** *<snp_filename>*
    	Argument used to define a file of SNP IDs to include.
    **--filter-exclude-snp-file** *<snp_filename>*
    	Argument used to define a file of SNP IDs to exclude.
    **--filter-max-missing** *proportion_float*
    	Argument used to filter positions by their proportion of missing data, a value of
        0.0 allows for no missing whereas a value of 1.0 ignores missing data. 
    **--filter-max-missing-count** *count_int*
        Argument used to filter positions by the number of samples with missing data, a 
        value of 0 allows for no samples to have missing data.

    ************************
    Position-Based Arguments
    ************************
    **--filter-include-pos** *<chr, chr:pos, chr:start-end, chr:start->*
        Argument used to include matching positions. May be used to include: an
        entire chromosome (i.e. chr); a single position (i.e. chr:pos); a 
        chromosomal locus (i.e. chr:start-end); or a chromosomal span (i.e.
        chr:start-/chr:0-end). This argument may be used multiple times if desired.
    **--filter-exclude-pos** *<chr, chr:pos, chr:start-end, chr:start->*
        Argument used to exclude matching positions. May be used to exclude: an
        entire chromosome (i.e. chr); a single position (i.e. chr:pos); a 
        chromosomal locus (i.e. chr:start-end); or a chromosomal span (i.e.
        chr:start-/chr:0-end). This argument may be used multiple times if desired.
    **--filter-include-pos-file** *<position_filename>*
        Argument used to define a file of positions to include within a tsv file 
        (chromosome and position).
    **--filter-exclude-pos-file** *<position_filename>*
        Argument used to define a file of positions to exclude within a tsv file 
        (chromosome and position).
    **--filter-include-bed** *<position_bed_filename>*
        Argument used to define a BED file of positions to include. Please note that
        filename must end in .bed.
    **--filter-exclude-bed** *<position_bed_filename>*
        Argument used to define a BED file of positions to exclude. Please note that
        filename must end in .bed.

    ********************
    Flag-Based Arguments
    ********************
    **--filter-include-passed**
        Argument used to include positions with the 'PASS' filter flag.
    **--filter-exclude-passed**
        Argument used to exclude positions with the 'PASS' filter flag.
    **--filter-include-filtered** *<filter_flag>*
        Argument used to include positions with the specified filter flag.
    **--filter-exclude-filtered** *<filter_flag>*
        Argument used to exclude positions with the specified filter flag.
'''

import os
import sys
import subprocess
import argparse
import logging

# Import basic bcftools functions
from pgpipe.bcftools import *

# Model file related functions
from pgpipe.model import read_model_file

# Import logging functions
from pgpipe.logging_module import initLogger, logArgs

def vcf_filter_parser(passed_arguments):
    '''
    VCF Filter Argument Parser

    Assign the parameters for the VCF Filter using argparse.

    Parameters
    ----------
	passed_arguments : list, optional
		Parameters passed by another function. sys.argv is used if
		not given. 

	Raises
    ------
    IOError
        If the input, or other specified files do not exist
    '''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def parser_add_to_list ():
        '''Custom action to add items to a list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):

                # Clean up any commas
                value = [item.strip(',') for item in value]
                
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    vcf_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input arguments.
    vcf_parser.add_argument('--vcf', help = "Defines the filename of the VCF", type = str, required = True, action = parser_confirm_file())

    # Model file arguments.
    vcf_parser.add_argument('--model-file', help = 'Defines the model file', type = str, action = parser_confirm_file())
    vcf_parser.add_argument('--model', help = 'Defines the model and the individual(s) to include', type = str)

    # Other file arguments. Expand as needed
    vcf_parser.add_argument('--out', help = 'Defines the complete output filename, overrides --out-prefix', type = str)
    vcf_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', type = str,  default = 'out')
    #vcf_parser.add_argument('--log', help = "Specifies if the vcftools log should be saved", action = 'store_false')

    out_format_list = ['vcf', 'vcf.gz', 'bcf', 'bed', 'sites']
    out_format_default = 'vcf.gz'

    vcf_parser.add_argument('--out-format', metavar = metavar_list(out_format_list), help = 'Defines the desired output format', type = str, choices = out_format_list, default = out_format_default)

    # General arguments.
    vcf_parser.add_argument('--overwrite', help = "Overwrite previous output", action = 'store_true')

    # Galaxy Option to pipe log to stdout
    vcf_parser.add_argument('--log-stdout', help = argparse.SUPPRESS, action = 'store_true')

    ### Filters

    # Individual list arguments
    indv_filters = vcf_parser.add_mutually_exclusive_group()
    indv_filters.add_argument('--filter-include-indv', help = 'Defines the individual(s) to include. May be used multiple times', nargs = '+', type = str, action = parser_add_to_list())
    indv_filters.add_argument('--filter-exclude-indv', help = 'Defines the individual(s) to exclude. May be used multiple times', nargs = '+', type = str, action = parser_add_to_list())
    
    # Individual file arguments
    indv_file_filters = vcf_parser.add_mutually_exclusive_group()
    indv_file_filters.add_argument('--filter-include-indv-file', help = 'Defines a file of individuals to include', action = parser_confirm_file())
    indv_file_filters.add_argument('--filter-exclude-indv-file', help = 'Defines a file of individuals to exclude', action = parser_confirm_file())

    # Allele count filters
    vcf_parser.add_argument('--filter-min-alleles', help = 'Include positions with a number of allele >= to the given number', type = int)
    vcf_parser.add_argument('--filter-max-alleles', help = 'Include positions with a number of allele <= to the given number', type = int)

    # Missing data filters
    missing_filters = vcf_parser.add_mutually_exclusive_group()
    missing_filters.add_argument('--filter-max-missing', help = 'Max proportion of missing data allowed (0.0: no missing data, 1.0: include all data)', type = float)
    missing_filters.add_argument('--filter-max-missing-count', help = 'Max number of sample with missing data allowed', type = int)

    # Indel variant-type filters
    indel_filters = vcf_parser.add_mutually_exclusive_group()
    indel_filters.add_argument('--filter-include-indels', help = 'Include positions if they contain an insertion or a deletion', action = 'store_true')
    indel_filters.add_argument('--filter-exclude-indels', help = 'Exclude positions if they contain an insertion or a deletion', action = 'store_true')
    
    # SNP variant-type filter
    snps_filters = vcf_parser.add_mutually_exclusive_group()
    snps_filters.add_argument('--filter-include-snps', help = 'Include positions if they contain a SNP', action = 'store_true')
    snps_filters.add_argument('--filter-exclude-snps', help = 'Exclude positions if they contain a SNP', action = 'store_true')

    # Position filters
    vcf_parser.add_argument('--filter-include-pos', help = 'Defines comma seperated positions (i.e. CHROM:START-END) to include. START and END are optional. May be used multiple times', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-exclude-pos', help = 'Defines comma seperated positions (i.e. CHROM:START-END) to exclude. START and END are optional. May be used multiple times', nargs = '+', type = str, action = parser_add_to_list())

    # Position file filters
    vcf_parser.add_argument('--filter-include-pos-file', help = 'Defines a file of positions to include within a file', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-pos-file', help = 'Defines a file of positions to exclude within a file', action = parser_confirm_file())

    # BED position filters
    vcf_parser.add_argument('--filter-include-bed', help = 'Defines a BED file of positions to include', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-bed', help = 'Defines a BED file of positions to exclude', action = parser_confirm_file())

    # Filter-flag filters
    vcf_parser.add_argument('--filter-include-passed', help = "Include positions with the 'PASS' filter flag", action = 'store_true')
    vcf_parser.add_argument('--filter-exclude-passed', help = "Exclude positions with the 'PASS' filter flag", action = 'store_true')
    vcf_parser.add_argument('--filter-include-flag', help = 'Include positions with the specified filter flag', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-exclude-flag', help = 'Exclude positions with the specified filter flag', nargs = '+', type = str, action = parser_add_to_list())

    # Info-flag filters
    #vcf_parser.add_argument('--filter-include-info', help = 'Include positions with the specified info flag', nargs = '+', type = str, action = parser_add_to_list())
    #vcf_parser.add_argument('--filter-exclude-info', help = 'Exclude positions with the specified info flag', nargs = '+', type = str, action = parser_add_to_list())

    # SNP filters
    vcf_parser.add_argument('--filter-include-snp', help = 'Include SNP(s) with the matching ID. This argument may be used multiple times', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-exclude-snp', help = 'Exclude SNP(s) with the matching ID. This argument may be used multiple times', nargs = '+', type = str, action = parser_add_to_list())

    # SNP file fitlers
    vcf_parser.add_argument('--filter-include-snp-file', help = 'Defines a file of SNP IDs to include', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-snp-file', help = 'Defines a file of SNP IDs to exclude', action = parser_confirm_file())

    # MAF Filter
    vcf_parser.add_argument('--filter-maf-min', help = 'Include sites with equal or greater MAF values', type = float)
    vcf_parser.add_argument('--filter-maf-max', help = 'Include sites with equal or lesser MAF values', type = float)

    # MAC Filter
    vcf_parser.add_argument('--filter-mac-min', help = 'Include sites with equal or greater MAC values', type = int)
    vcf_parser.add_argument('--filter-mac-max', help = 'Include sites with equal or lesser MAC values', type = int)

    if passed_arguments:
        return vcf_parser.parse_args(passed_arguments)
    else:
        return vcf_parser.parse_args()

def run (passed_arguments = []):
    '''
    Filter function for the PPP

    This function uses the argparse-based function :py:func:`vcf_filter_parser`
    to parse either sys.argv or passed_arguments to obtain the parameters below. 
    The parameters are then translated to their bcftools equivalent. Once all the
    bcftools-based parameters are assigned, bcftools is called.

    Parameters
    ----------
    --vcf : str
        Input VCF filename
    --out : str
        Complete output filename, overrides --out-prefix
    --out-prefix : str
        Output filename prefix
    --out-format : str
        Output format
	--model-file : str
		Model filename
	--model : str
		Model to use
    --filter-include-indv : list or str
        Individual(s) to include. May be used multiple times. Not usable w/ --model
    --filter-exclude-indv : list or str
        Individual(s) to exclude. May be used multiple times. Not usable w/ --model
    --filter-include-indv-file : str
        File of individuals to include. Not usable w/ --model
    --filter-exclude-indv-file : str
        File of individuals to exclude.  Not usable w/ --model
    --filter-min-alleles : int
        Include positions with a number of allele >= to the given number
    --filter-max-alleles : int
        Include positions with a number of allele <= to the given number
    --filter-maf-min
		Include sites with equal or greater MAF values
    --filter-maf-max
    	Include sites with equal or lesser MAF values
    --filter-mac-min
    	Include sites with equal or greater MAC values
    --filter-mac-max
    	Include sites with equal or lesser MAC values
   	--filter-include-indels
    	Include positions if they contain an insertion or a deletion
    --filter-exclude-indels
    	Exclude positions if they contain an insertion or a deletion
 	--filter-include-snp
 		Include SNP(s) with the matching ID. May be used multiple times
    --filter-include-snp-file
    	File of SNP IDs to include
    --filter-exclude-snp-file
    	File of SNP IDs to exclude
    --filter-max-missing
		Filter positions by their missing data proportion. Values should be given
        as max proportion of missing data to accept. For example, a value of 0.0
        allows no missing data whereas a value of 0.1 allows for 10% of samples to 
        have missig data.
    --filter-include-chr
        Chromosome(s) to include. May be used multiple times.
    --filter-exclude-chr
        Chromosome(s) to exclude. May be used multiple times.
    --filter-from-bp : int
        Lower bound of positions to include. Only usable with a single chromosome
    --filter-to-bp : int
        Upper bound of positions to include. Only usable with a single chromosome
    --filter-include-positions : str
        File of positions to include (tsv: chromosome and position)
    --filter-exclude-positions : str
        File of positions to exclude (tsv: chromosome and position)
    --filter-include-bed : str
        BED file of positions to include 
    --filter-exclude-bed : str
        BED file of positions to exclude
    --filter-include-passed : bool
        Include positions with the 'PASS' filter flag
    --filter-include-filtered : list or str
        Include positions with the specified filter flag
    --filter-exclude-filtered : list or str
        Exclude positions with the specified filter flag
    --filter-include-info : list or str
        Include positions with the specified info flag
    --filter-exclude-info : list or str
        Exclude positions with the specified info flag

    Raises
    ------
    IOError
        Output file already exists and --overwrite is not specified
    Exception
        Incompatible arguments

    '''

    # Grab VCF arguments from command line
    vcf_args = vcf_filter_parser(passed_arguments)

    # Check if the input file is indexed
    vcf_is_indexed = check_for_index(vcf_args.vcf)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(vcf_args, func_name = 'vcf_filter')

    # Argument container for bcftools
    bcftools_call_args = ['view']

    # Check if the user has specified a model file
    if vcf_args.model_file and vcf_args.model:
        # Read in the models
        models_in_file = read_model_file(vcf_args.model_file)

        # Check that the selected model was not found in the file
        if vcf_args.model not in models_in_file:
            raise IOError('Selected model "%s" not found in: %s' % (vcf_args.model, vcf_args.model_file))

        # Check that individual-based filters are not also being used
        if vcf_args.filter_include_indv or vcf_args.filter_exclude_indv:
            if vcf_args.filter_include_indv:
                raise Exception('--model and --filter-include-indv arguments are incompatible')
            if vcf_args.filter_exclude_indv:
                raise Exception('--model and --filter-exclude-indv arguments are incompatible')

        # Check that individuals-based filters are not also being used
        if vcf_args.filter_include_indv_file or vcf_args.filter_exclude_indv_file:
            if vcf_args.filter_include_indv_file:
                raise Exception('--model and --filter_include_indv-file arguments are incompatible')
            if vcf_args.filter_exclude_indv_file:
                raise Exception('--model and --filter-exclude-indv-file arguments are incompatible')

        # Select model, might change this in future versions
        selected_model = models_in_file[vcf_args.model]

        # Create individuals file
        selected_model.create_ind_file(overwrite = vcf_args.overwrite)

        # Assign the individuals file to bcftools
        bcftools_call_args.extend(['--samples-file', selected_model.ind_file])

    # Holds the filename suffix bcftools assigns to the filtered output
    bcftools_out_suffix = None

    # Assign expected bcftools output filename
    bcftools_output_filename = '%s.%s' % (vcf_args.out_prefix, vcf_args.out_format)

    # Check if the user has specified a output filename
    if vcf_args.out:

        # Rename the bcftools output filename, using the output filename
        bcftools_output_filename = vcf_args.out

    # Check if previous output should be overwritten
    if not vcf_args.overwrite:

        # Confirm the bcftools output and log file do not exist
        check_for_bcftools_output(bcftools_output_filename)

    # Individuals-based filters
    if vcf_args.filter_include_indv_file or vcf_args.filter_exclude_indv_file:
        # Used to include a file of individuals to keep
        if vcf_args.filter_include_indv_file:
            bcftools_call_args.extend(['--samples-file', vcf_args.filter_include_indv_file])

        # Used to include a file of individuals to remove
        if vcf_args.filter_exclude_indv_file:
            bcftools_call_args.extend(['--samples-file', '^' + vcf_args.filter_exclude_indv_file])

    # Individual-based filters
    if vcf_args.filter_include_indv or vcf_args.filter_exclude_indv:
        if vcf_args.filter_include_indv:
            bcftools_call_args.extend(['--samples', ','.join(vcf_args.filter_include_indv)])
        if vcf_args.filter_exclude_indv:
            bcftools_call_args.extend(['--samples', '^' + ','.join(vcf_args.filter_exclude_indv)])

    # Position-list filter
    if vcf_args.filter_include_pos or vcf_args.filter_exclude_pos:

        # Check if positions are to be included
        if vcf_args.filter_include_pos:
            # Check if the input is indexed
            if vcf_is_indexed == True:
                bcftools_call_args.extend(['--regions', ','.join(vcf_args.filter_include_pos)])

            # Check if the input is not indexed
            elif vcf_is_indexed != True:
                bcftools_call_args.extend(['--targets', ','.join(vcf_args.filter_include_pos)])
        
        # Check if positions are to be excluded
        if vcf_args.filter_exclude_pos:
            bcftools_call_args.extend(['--targets', '^' + ','.join(vcf_args.filter_exclude_pos)])

    # Position file filters that do not require indexed input
    if (vcf_args.filter_include_pos_file or vcf_args.filter_exclude_pos_file) and vcf_is_indexed != True:
        if vcf_args.filter_include_pos_file:
            bcftools_call_args.extend(['--targets-file', vcf_args.filter_include_pos_file])
        if vcf_args.filter_exclude_pos_file:
            bcftools_call_args.extend(['--targets-file', '^' + vcf_args.filter_exclude_pos_file])

    # Position file filters that requires indexed input
    if (vcf_args.filter_include_pos_file or vcf_args.filter_exclude_pos_file) and vcf_is_indexed == True:
        if vcf_args.filter_include_pos_file:
            bcftools_call_args.extend(['--regions-file', vcf_args.filter_include_pos_file])
        if vcf_args.filter_exclude_pos_file:
            bcftools_call_args.extend(['--targets-file', '^' + vcf_args.filter_exclude_pos_file])

    # Position BED filters that do not require indexed input
    if (vcf_args.filter_include_bed or vcf_args.filter_exclude_bed) and vcf_is_indexed != True:
        if vcf_args.filter_include_bed:
            bcftools_call_args.extend(['--targets-file', vcf_args.filter_include_bed])
        if vcf_args.filter_exclude_bed:
            bcftools_call_args.extend(['--targets-file', '^' + vcf_args.filter_exclude_bed])

    # Position BED filters that requires indexed input
    if (vcf_args.filter_include_bed or vcf_args.filter_exclude_bed) and vcf_is_indexed == True:
        if vcf_args.filter_include_bed:
            bcftools_call_args.extend(['--regions-file', vcf_args.filter_include_bed])
        if vcf_args.filter_exclude_bed:
            bcftools_call_args.extend(['--targets-file', '^' + vcf_args.filter_exclude_bed])

    # Include variants (indel, snp, or both)
    if vcf_args.filter_include_indels or vcf_args.filter_include_snps:

        # List to store variants to include
        include_variant_types = []

        # Check if indels should be included
        if vcf_args.filter_include_indels:
            
            # Add the indels type
            include_variant_types.append('indels')

        # Check if SNPs should be included
        if vcf_args.filter_include_snps:
            
            # Add the snps type
            include_variant_types.append('snps')

        # Store the types to include argument
        bcftools_call_args.extend(['--types', '%s' % ','.join(include_variant_types)])

    # Exclude variants (indel, snp, or both)
    if vcf_args.filter_exclude_indels or vcf_args.filter_exclude_snps:

        # List to store variants to exclude
        exclude_variant_types = []

        # Check if indels should be excluded
        if vcf_args.filter_exclude_indels:
            
            # Add the indels type
            exclude_variant_types.append('indels')

        # Check if SNPs should be excluded
        if vcf_args.filter_exclude_snps:
            
            # Add the snps type
            exclude_variant_types.append('snps')

        # Store the types to include argument
        bcftools_call_args.extend(['--exclude-types', '%s' % ','.join(exclude_variant_types)])

    # Allele-based filters
    if vcf_args.filter_min_alleles or vcf_args.filter_max_alleles:
        if vcf_args.filter_min_alleles:
            bcftools_call_args.extend(['--min-alleles', vcf_args.filter_min_alleles])
        if vcf_args.filter_max_alleles:
            bcftools_call_args.extend(['--max-alleles', vcf_args.filter_max_alleles])

    # MAF-based filters
    if vcf_args.filter_maf_min != None or vcf_args.filter_maf_max != None:
        if vcf_args.filter_maf_min != None:
            bcftools_call_args.extend(['--min-af', vcf_args.filter_maf_min])
        if vcf_args.filter_maf_max != None:
            bcftools_call_args.extend(['--max-af', vcf_args.filter_maf_max])

    # MAC-based filters
    if vcf_args.filter_mac_min != None or vcf_args.filter_mac_max != None:
        if vcf_args.filter_mac_min != None:
            bcftools_call_args.extend(['--min-ac', vcf_args.filter_mac_min])
        if vcf_args.filter_mac_max != None:
            bcftools_call_args.extend(['--max-ac', vcf_args.filter_mac_max])

    # List to store expression for BCFtools include expression functionality
    bcftools_expressions = []

    # SNP filters
    if vcf_args.filter_include_snp or vcf_args.filter_exclude_snp:

        # Check for SNPs to include
        if vcf_args.filter_include_snp:

            # Create list to store include SNP expression
            include_snp_expression_list = []

            # Loop the SNPs
            for snp_to_include in vcf_args.filter_include_snp:
                
                # Save the include SNP expression       
                include_snp_expression_list.append('%%ID=="%s"' % snp_to_include)

            # Merge the include SNP expression
            include_snp_expression_str = '%s' % ' || '.join(include_snp_expression_list)

            # Check if parentheses are needed
            if len(include_snp_expression_list) > 1:

                # Add parentheses
                include_snp_expression_str = '(%s)' % include_snp_expression_str

            # Add the include expression to the list
            bcftools_expressions.append(include_snp_expression_str)

        # Check for SNPs to exclude
        if vcf_args.filter_exclude_snp:

            # Create list to store exclude SNP expression
            exclude_snp_expression_list = []

            # Loop the SNPs
            for snp_to_exlude in vcf_args.filter_exclude_snp:
                
                # Save the include SNP expression       
                exclude_snp_expression_list.append('%%ID!="%s"' % snp_to_exlude)

            # Merge the exclude SNP expression
            exclude_snp_expression_str = '%s' % ' && '.join(exclude_snp_expression_list)

            # Check if parentheses are needed
            if len(exclude_snp_expression_list) > 1:

                # Add parentheses
                exclude_snp_expression_str = '(%s)' % exclude_snp_expression_str

            # Add the exclude expression to the list
            bcftools_expressions.append(exclude_snp_expression_str)

    # SNP-file filters
    if vcf_args.filter_include_snp_file or vcf_args.filter_exclude_snp_file:
        if vcf_args.filter_include_snp_file:
            bcftools_expressions.append('%%ID==@%s' % vcf_args.filter_include_snp_file)
        if vcf_args.filter_exclude_snp_file:
            bcftools_expressions.append('%%ID!=@%s' % vcf_args.filter_exclude_snp_file)

    # Missing data filters
    if vcf_args.filter_max_missing != None or vcf_args.filter_max_missing_count != None:
        if vcf_args.filter_max_missing != None:
            bcftools_expressions.append('F_MISSING <= %s' % vcf_args.filter_max_missing)
        if vcf_args.filter_max_missing_count  != None:
            bcftools_expressions.append('N_MISSING <= %s' % vcf_args.filter_max_missing_count)

    # PASS flag Filters
    if vcf_args.filter_include_passed or vcf_args.filter_exclude_passed:

        if vcf_args.filter_include_passed:
            bcftools_expressions.append('%FILTER=="PASS"')

        if vcf_args.filter_exclude_passed:
            bcftools_expressions.append('%FILTER!="PASS"')

    # General Flag filters
    if vcf_args.filter_include_flag or vcf_args.filter_exclude_flag:

        # Check for flags to include
        if vcf_args.filter_include_flag:

            # List to store variants to include
            include_flag_list = []

            # Loop flags to include
            for filtered_to_include in vcf_args.filter_include_flag:
                
                # Save the flag to include
                include_flag_list.append('%%FILTER=="%s"' % filtered_to_include)

            # Merge the flags to include
            include_flag_str = '%s' % ' || '.join(include_flag_list)

            # Check if parentheses are needed
            if len(include_flag_list) > 1:

                # Add parentheses
                include_flag_str = '(%s)' % include_flag_str

            # Append the include flag expression
            bcftools_expressions.append(include_flag_str)

        # Check for flags to exclude
        if vcf_args.filter_exclude_flag:

            # List to store variants to exclude
            exclude_flag_list = []

            # Loop flags to exclude
            for filtered_to_exclude in vcf_args.filter_exclude_flag:
                
                # Save the flag to exclude
                exclude_flag_list.append('%%FILTER!="%s"' % filtered_to_exclude)

            # Merge the flags to exclude
            exclude_flag_str = '%s' % ' && '.join(exclude_flag_list)

            # Check if parentheses are needed
            if len(exclude_flag_list) > 1:

                # Add parentheses
                exclude_flag_str = '(%s)' % exclude_flag_str

            # Append the exclude flag expression
            bcftools_expressions.append(exclude_flag_str)

    '''
    # Info-based filters
    if vcf_args.filter_include_info or vcf_args.filter_exclude_info:
        if vcf_args.filter_include_info:
            for info_to_include in vcf_args.filter_include_info:
                bcftools_call_args.extend(['--keep-INFO', info_to_include])
        if vcf_args.filter_exclude_info:
            for info_to_exclude in vcf_args.filter_exclude_info:
                bcftools_call_args.extend(['--remove-INFO', info_to_exclude])
    '''

    # Save the BCFtools expression argument
    bcftools_expression_arg = '(%s)' % ' && '.join(bcftools_expressions)

    # Add the expression argument to the argument list
    bcftools_call_args.extend(['--include', bcftools_expression_arg])

    logging.info('bcftools parameters assigned')

    # Assign the input file
    bcftools_call_args.append(vcf_args.vcf)

    logging.info('Input file assigned')

    if vcf_args.out_format in ['vcf', 'vcf.gz', 'bcf']:

        # Assign the output format
        bcftools_call_args.extend(return_output_format_args(vcf_args.out_format))

        # Assign the output filename
        bcftools_call_args.extend(['-o', bcftools_output_filename])

        # Call bcftools with the specifed arguments
        bcftools_err = call_bcftools(bcftools_call_args)

    elif vcf_args.out_format in ['bed', 'sites']:

        # Create list for the conversion call
        bcftools_cvt_args = ['query', '-f']

        # Check if the output format is bed
        if vcf_args.out_format == 'bed':

            # Add the arguments to create a bed file
            bcftools_cvt_args.append('%CHROM\t%POS0\t%END\t%ID\n')

        # Check if the output format is sites
        elif vcf_args.out_format == 'sites':

            # Add the arguments to create a bed file
            bcftools_cvt_args.append('%CHROM\t%POS\n')

        # Assign the output filename
        bcftools_cvt_args.extend(['-o', bcftools_output_filename])

        print(bcftools_call_args)

        # Call BCFtools twice
        pipe_bcftools_bcftools(bcftools_call_args, bcftools_cvt_args)

    else:
        raise Exception('Unknown file format: %s' % vcf_args.out_format)

    # Check if the log should be piped to the stdout
    if vcf_args.log_stdout:

        # Write the log to stdout
        stdout_bcftools_reference()

    # Check if log should be saved as a file
    else:

        # Create the log file
        log_bcftools_reference(bcftools_output_filename)

    # Delete any files that were created for bcftools
    if vcf_args.model_file:
        selected_model.delete_ind_file()

if __name__ == "__main__":
    initLogger()
    run()



import os
import sys
import gzip
import logging
from Bio import SeqIO

def check_fasta_gz (fasta_filename):

	try:
		fasta_file = gzip.open(fasta_filename)
		fasta_line = fasta_file.readline()
		fasta_file.close()

		if fasta_line[:1] == b'>':
			return True

		raise Exception('%s is not fasta formatted' % fasta_filename)

	except:

		return False

def check_fasta (fasta_filename):

	fasta_file = open(fasta_filename)
	fasta_line = fasta_file.readline()
	fasta_file.close()

	if fasta_line[:1] == b'>':
		return True

	raise Exception('%s is not fasta formatted' % fasta_filename)

def check_format(fasta_filename):
    """Checks header of given file for compression type


    Given a filename, opens file and reads first line to check if
    file has BGZF or GZIP header. May be extended to check for BCF format

    Parameters
    ----------
    fasta_filename : str
        Name of file to be checked

    Returns
    -------
    extension : str {'gzip', 'fasta', 'other'}
        File extension as indicated by header

    """

    # Check if the file is bgzipped or gunzipped, both of which should function
    if check_fasta_gz(fasta_filename):
   		return 'gzip'

    # Check if the file is in the fasta format
    if check_fasta(fasta_filename):
    	return 'fasta'
